//////////////////////////////////////////////////////////////////////////
//	"getopt.cpp"
//------------------------------------------------------------------------
//	$Id: getopt.cpp 4 2006-03-01 22:26:13Z taro $
//////////////////////////////////////////////////////////////////////////

#include	"getopt.h"
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>

#if defined(WIN32)
# include	<crtdbg.h>
# define	FI64	"I64"
# define	I64(n)	n##i64
#elif defined(LINUX)
# include	<assert.h>
# define	_ASSERT
# define	FI64	"ll"
# define	I64(n)	n##ll
# define	_istdigit	isdigit
# define	_putts		puts
# define	_tcslen		strlen
# define	_tcscpy		strcpy
# define	_tcsicmp	strcasecmp
# define	_tcsnicmp	strncasecmp
# define	_tcsncpy	strncpy
# define	_totupper	toupper
#endif

#if	0 // defined(_DEBUG)
# define	DBGPRINT(param_)	printf param_
#else
# define	DBGPRINT(param_)
#endif


GetOpt::GetOpt()
{
	this->last_pos = this->cur_pos = this->next_pos = this->swpos = -1;
	this->last_param = NULL;
	this->err_option = NULL;
	this->err_param = NULL;

	this->coltbl[COL_1] =  2;
	this->coltbl[COL_2] =  6;
	this->coltbl[COL_3] = 26;
	this->coltbl[COL_4] = 29;
}

GetOpt::~GetOpt()
{
	{
		std::list<FuncItem *>::iterator	it;

		for ( it = this->funcList.begin(); it != this->funcList.end(); it++ )
			delete *it;
	}
	this->clear();
}

int		GetOpt::AddFunc(GetOpt::GetVarFunc func)
{
	FuncItem	*item;

	if ( (item = new FuncItem) == NULL )
		return -1;
	item->vartype = (int)VAR_FUNC + (int)this->funcList.size();
	item->func = func;
	this->funcList.push_back(item);
	return item->vartype;
}

GetOpt::GetVarFunc	GetOpt::GetFunc(int vartype)
{
	FuncItem						*item = NULL;
	std::list<FuncItem *>::iterator	it;

	for ( it = this->funcList.begin(); it != this->funcList.end(); it++ )
	{
		item = *it;

		if ( item->vartype == vartype )
			break;
	}
	return item != NULL ? item->func : NULL;
}

GetOpt::Item *	GetOpt::AddItem(int usertype, const TCHAR * name, const _TCHAR * sw, VarType vartype)
{
	Item	*item;

	DBGPRINT(("GetOpt::AddItem(usertype=%d, name=%s\n", usertype, name));
	if ( (item = new Item) == NULL )
		return NULL;
	item->usertype    = usertype;
	item->name        = name;
	item->name_len    = name != NULL ? (unsigned)_tcslen(name) : 0;
	item->sw          = sw;
	item->vartype     = vartype;
	item->sel.seltype = SEL_OFF;
	item->sel.tbl     = NULL;
	item->std.vartype = VAR_NOTHING;
	item->var.vartype = VAR_NOTHING;
	item->ptr.vartype = VAR_NOTHING;
	item->ptr.text    = NULL;

	this->push_back(item);

	return item;
}

int		GetOpt::AddItem(unsigned n, GetOpt::AddItemParam * tbl)
{
	unsigned	i;

	for ( i = 0; i < n; i++ )
	{
		AddItemParam	*param;
		Item			*item;

		param = &tbl[i];
		if ( (item = this->AddItem(param->usertype, param->name, param->sw, param->vartype)) == NULL )
			return -1;	// error

		switch ( param->vartype )
		{
		case VAR_FLAG:
			item->std.vartype = VAR_FLAG;
			item->std.flag    = param->omit;
			break;

		case VAR_TEXT:
			if ( param->omit )
			{
				item->std.vartype = VAR_TEXT;
				item->std.text    = param->std.text;
			}
			break;

		case VAR_INT:
			if ( param->omit )
			{
				item->std.vartype = VAR_INT;
				item->std.n       = param->std.n;
			}
			break;

		case VAR_INT64:
			if ( param->omit )
			{
				item->std.vartype = VAR_INT64;
				item->std.n64     = param->std.n64;
			}
			break;
		}
	}

	return 0;
}

bool	GetOpt::IsSwitch(const _TCHAR *s)
{
	return  *s == _TEXT('-') ? true : false;
}

const _TCHAR *	GetOpt::GetCurArg(int argc, const _TCHAR * argv[])
{
	if ( this->cur_pos >= 0 && this->cur_pos < argc )
		return argv[this->cur_pos];
	else
		return NULL;
}

GetOpt::Item *	GetOpt::GetItem(int usertype, const _TCHAR * name)
{
	Item		*item;
	iterator	it;

	for ( it = this->begin(); it != this->end(); it++ )
	{
		item = *it;
		if ( item->usertype == usertype 
		  && (name == NULL || _tcsicmp(name, item->name) == 0) )
		{
			break;
		}
	}
	return item;
}

int	GetOpt::GetVar(GetOpt::Item * item, const _TCHAR * s)
{
	int		retcode = -1;

	switch ( item->vartype )
	{
	case VAR_NOTHING:
		return 0;

	case VAR_FLAG:
		if ( s == NULL || *s == _TEXT('\0') )
		{
			if ( item->std.vartype == VAR_FLAG )
			{	// set default value
				item->var = item->std;
				retcode = 0;	// not use param
			} else
				return -1;	// error
		} else if ( _tcsicmp(s, _TEXT("on")) == 0 )
		{
			item->var.vartype = VAR_FLAG;
			item->var.flag = true;
			retcode = 1;
		} else if ( _tcsicmp(s, _TEXT("off")) == 0 )
		{
			item->var.vartype = VAR_FLAG;
			item->var.flag = false;
			retcode = 1;
		} else
		{
			if ( item->std.vartype == item->vartype && this->IsSwitch(s) )
			{	// set default value
				item->var = item->std;
				retcode = 0;	// not use param
			}
		}
		if ( item->ptr.vartype == VAR_FLAG )
			*(item->ptr.flag) = item->var.flag;
		return 0;

	case VAR_TEXT:
	{
		if ( s == NULL || *s == _TEXT('\0') )
		{
			if ( item->std.vartype == item->vartype )
			{	// set default value
				item->var = item->std;
				retcode = 0;	// not use param
			} else
				return -1;	// error
		} else
		{
			if ( item->std.vartype == item->vartype && this->IsSwitch(s) )
			{	// ready default value & next switch
				item->var = item->std;
				retcode = 0;	// not use param
			} else
			{
				item->var.vartype = VAR_TEXT;
				item->var.text = s;
				retcode = 1;	// use param
			}
		}
		switch ( item->sel.seltype )
		{
		case SEL_TABLE:
		{
			int				i;

			_ASSERT(item->sel.tblnum > 0 && item->sel.tbl != NULL );
			for ( i = 0; i < item->sel.tblnum; i++ )
			{
				if ( item->sel.tbl[i].vartype == VAR_TEXT
				  && item->sel.tbl[i].text != NULL
				  && _tcsicmp(item->sel.tbl[i].text, item->var.text) == 0 )
				{
					item->var.selno = item->sel.tbl[i].selno;
					break;	// OK
				}
			}
			if ( i >= item->sel.tblnum )
			{
				this->err_param = s;
				return -1;	// select error
			}
			break;
		}

		default:
			break;
		}
		if ( item->ptr.vartype == VAR_TEXT )
			*(item->ptr.text) = item->var.text;
		break;
	}

	case VAR_INT:
	case VAR_INT64:
		if ( s == NULL || *s == _TEXT('\0') )
		{
			if ( item->std.vartype == item->vartype )
			{	// set default value
				item->var = item->std;
				retcode = 0;	// not use param
			} else
				return -1;	// error
		} else
		{
			if ( _istdigit(*s)
			  || ((*s == _TEXT('+') || *s == _TEXT('-')) && _istdigit(s[1]))
			)
			{
				item->var.vartype = item->vartype;
				switch ( item->vartype )
				{
				case VAR_INT  : item->var.n   = (int) GetOpt::GetInt64(s); break;
				case VAR_INT64: item->var.n64 =       GetOpt::GetInt64(s); break;
				default: _ASSERT(0); return -1;
				}
				retcode = 1;	// use param
			} else if ( this->IsSwitch(s) && item->std.vartype == item->vartype )
			{	// set default value
				item->var = item->std;
				retcode = 0;	// not use param
			} else
				return -1;	// error
		}
		switch ( item->sel.seltype )
		{
		case SEL_TABLE:
		{
			int				i;

			_ASSERT(item->sel.tblnum > 0 && item->sel.tbl != NULL );
			for ( i = 0; i < item->sel.tblnum; i++ )
			{
				if ( item->sel.tbl[i].vartype == VAR_INT
				  && item->sel.tbl[i].n == item->var.n )
				{
					item->var.selno = item->sel.tbl[i].selno;
					break;	// OK
				}
				if ( item->sel.tbl[i].vartype == VAR_INT64
				  && item->sel.tbl[i].n64 == item->var.n64 )
				{
					item->var.selno = item->sel.tbl[i].selno;
					break;	// OK
				}
			}
			if ( i >= item->sel.tblnum )
			{
				this->err_param = s;
				return -1;	// select error
			}
			break;
		}

		case SEL_MINMAX:
			_ASSERT(item->sel.tblnum >= 2 && item->sel.tbl != NULL );
			switch ( item->vartype )
			{
			case VAR_INT:
				if ( item->var.n < item->sel.tbl[0].n
				  && item->var.n > item->sel.tbl[1].n )
				{
					this->err_param = s;
					return -1; // min/max error
				}
				break;
			case VAR_INT64:
				if ( item->var.n64 < item->sel.tbl[0].n64
				  && item->var.n64 > item->sel.tbl[1].n64 )
				{
					this->err_param = s;
					return -1; // min/max error
				}
				break;
			default:
				_ASSERT(0);
				return -1;
			}
			break;
		}
		switch ( item->vartype )
		{
		case VAR_INT  : 
			if ( item->ptr.vartype == VAR_INT )
				*(item->ptr.n) = item->var.n;
			break;
		case VAR_INT64: 
			if ( item->ptr.vartype == VAR_INT64 )
				*(item->ptr.n64) = item->var.n64;
			break;
		default:
			 _ASSERT(0);
			return -1;
		}
		break;

	default:
	{
		GetVarFunc	func;

		if ( (func = this->GetFunc(item->vartype)) == NULL )
			return -1;
	//	return (this->*func)(item, s);
		return (*func)(item, s);
	}
	} // end of switch

	return retcode;
}

GetOpt::Item *	GetOpt::ParseArg(int & pos, int argc, const _TCHAR * argv[])
{
	int					res;
	const _TCHAR		*s;

	DBGPRINT(("GetOpt::ParseArg(%d, %d, 0x%08X)\n", pos, argc, (unsigned)argv));
	this->err_option = NULL;
	this->err_param  = NULL;
	this->last_param = NULL;
	if ( this->last_pos == pos && this->swpos > 0 )
		return this->ParseSwitch(pos, argc, argv);
	this->swpos   = -1;
	this->cur_pos = pos;
	if ( this->cur_pos < 0 )
		this->cur_pos = 0;
	else if ( this->cur_pos >= argc )
	{
		pos = this->cur_pos;
		return NULL;
	}

	this->last_pos = this->cur_pos;
	s = argv[this->cur_pos];
	if ( *s == _TEXT('-') )
	{
		++s;
		if ( *s == _TEXT('-') )
		{	// long option name
			GetOpt::iterator	it;

			++s;
			for ( it = this->begin(); it != this->end(); it++ )
			{
				Item	*item;

				item = *it;
				if ( item->name != NULL )
				{
					if ( _tcsicmp(item->name, s) == 0 )
					{
						this->cur_pos++;
						if ( this->cur_pos < argc )
							s = argv[this->cur_pos];
						else
							s = NULL;
						if ( (res = this->GetVar(item, s)) < 0 )
						{
							this->err_option = argv[this->last_pos];
							pos = this->cur_pos;
							return NULL;	// error
						}
						if ( res > 0 )
							this->cur_pos++;
						pos = this->cur_pos;
						return item;
					} else if ( _tcsnicmp(item->name, s, item->name_len) == 0 && s[item->name_len] == _TEXT('=') )
					{
						this->cur_pos++;
						if ( this->GetVar(item, &s[item->name_len+1]) < 0 )
						{
							this->err_option = argv[this->last_pos];
							pos = this->cur_pos;
							return NULL;
						}
						pos = this->cur_pos;
						return item;
					}
				}
			}
			this->err_option = argv[this->last_pos];
			pos = this->cur_pos;
			return NULL;	// unknown option
		} else
		{	// short option switch
			this->swpos = 1;
			return this->ParseSwitch(pos, argc, argv);
		}
		return NULL;
	}
	// no option

	this->last_param = argv[this->last_pos];
	this->cur_pos++;
	pos = this->cur_pos;
	return NULL;
}

// -abcde
// -c=test
// -b2048
GetOpt::Item *	GetOpt::ParseSwitch(int & pos, int argc, const _TCHAR * argv[])
{
	const _TCHAR		*s;
	int					ch;
	GetOpt::iterator	it;

	s = argv[this->cur_pos];
	ch = s[this->swpos++];
	for ( it = this->begin(); it != this->end(); it++ )
	{
		Item	*item;

		item = *it;
		if ( item->sw != NULL )
		{
			const _TCHAR	*sw;

			for ( sw = item->sw; *sw != _TEXT('\0'); ++sw )
			{
				if ( _totupper(*sw) == _totupper(ch) )
				{
					if ( s[this->swpos] == _TEXT('=') && this->swpos == 2 )
					{	// ex. -c=<param>
						this->swpos++;
						this->cur_pos++;
						if ( this->GetVar(item, &s[this->swpos]) < 0 )
						{
							this->err_option = argv[this->last_pos];
							pos = this->cur_pos;
							return NULL;
						}
						pos = this->cur_pos;
						return item;
					} else
					{
						if ( item->vartype != item->std.vartype )
						{	// no setup default value
							if ( this->swpos == 2 )
							{	// ex. -B2048
								if ( this->GetVar(item, &s[this->swpos]) < 0 )
								{
									this->err_option = argv[this->last_pos];
									this->err_param  = &s[this->swpos];
									pos = this->cur_pos;
									return NULL;
								}
								this->cur_pos++;
								this->swpos = -1;
								pos = this->cur_pos;
								return item;
							} else
							{
								this->err_option = argv[this->last_pos];
								this->err_param  = s;
								pos = this->cur_pos;
								return NULL;
							}
						}
						if ( this->GetVar(item, NULL) < 0 )
						{
							_ASSERT(0);
							this->err_option = argv[this->last_pos];
							this->err_param  = &s[this->swpos];
							pos = this->cur_pos;
							return NULL;
						}
						if ( s[this->swpos] == _TEXT('\0') )
						{
							this->cur_pos++;
							this->swpos = -1;
						}
						pos = this->cur_pos;
						return item;
					}
				}
			}
		}
	} // end of for(;;)
	// unknown switch
	this->err_option = argv[this->last_pos];
	this->err_param  = &s[this->swpos-1];
	return NULL;
}

GetOpt::Int64		GetOpt::GetInt64(const _TCHAR * s) // static
{
	Int64		n = 0;
	int			sign = 1;	

	_ASSERT(s != NULL);
	if ( *s == _TEXT('-') )
	{
		sign = -1;
		s++;
	}
	while ( *s )
	{
		if ( *s < _TEXT('0') || *s > _TEXT('9') )
			break;
		n = n * I64(10) + (Int64)(*s - _TEXT('0'));
		s++;
	}

	switch ( *s )
	{
		case _TEXT('K'):
			n = n * I64(0x400);
			break;
		case _TEXT('M'):
			n = n * I64(0x100000);
			break;
		case _TEXT('G'):
			n = n * I64(0x40000000);
			break;

		case _TEXT('k'):
			n = n * I64(1000);
			break;
		case _TEXT('m'):
			n = n * I64(1000000);
			break;
		case _TEXT('g'):
			n = n * I64(1000000000);
			break;
	}
	return sign < 0 ? -n : n;
}

void	GetOpt::Help(unsigned tblnum, const GetOpt::HelpItem * tbl)
{
	unsigned		i;

	for ( i = 0; i < tblnum; i++)
	{
		GetOpt::Item	* item;
		enum	Column
		{
			COL_1ST =  2,
			COL_2ND =  6,
			COL_3RD = 28,
		};

		if ( (item = tbl[i].item) == NULL )
			item = this->GetItem(tbl[i].usertype, NULL);
		if ( item != NULL )
		{
			_TCHAR				buffer[1024], *d;
			const _TCHAR		*s;
			int					c;

			c = 0;
			d = buffer;
			if ( item->sw != NULL && item->sw[0] != _TEXT('\0') )
			{
				for ( s = item->sw; *s != _TEXT('\0'); s++ )
				{
					if ( c > this->coltbl[COL_1] )
					{
						++c; *d++ = _TEXT(',');
						++c; *d++ = _TEXT(' ');
					} else if ( c >= 0 && c < this->coltbl[COL_1] )
					{
						while ( c < this->coltbl[COL_1] )
							c++, *d++ = _TEXT(' ');
					}
					c++; *d++ = _TEXT('-');
					c++; *d++ = *s;
				}
			}
			if ( item->name != NULL && item->name[0] != _TEXT('\0') )
			{
				unsigned	name_len;

				if ( d != buffer )
				{
					c++; *d++ = _TEXT(',');
					c++; *d++ = _TEXT(' ');
				}
				if ( c >= 0 && c < this->coltbl[COL_2] )
				{
					while ( c < this->coltbl[COL_2] )
						c++, *d++ = _TEXT(' ');
				}
				name_len = (unsigned)_tcslen(item->name);
				c++; *d++ = _TEXT('-');
				c++; *d++ = _TEXT('-');
				_tcsncpy(d, item->name, name_len);
				d += name_len;
				c += name_len;
			}
			if ( tbl[i].param != NULL && tbl[i].param[0] != _TEXT('\0') )
			{
				unsigned	param_len;

				c++; *d++ = _TEXT(' ');
				param_len = (unsigned)_tcslen(tbl[i].param);
				_tcsncpy(d, tbl[i].param, param_len);
				d += param_len;
				c += param_len;
			}
			if ( c >= this->coltbl[COL_3] )
			{
				c++; *d++ = _TEXT('\n');
				c = 0;
			}
			while ( c < this->coltbl[COL_3] )
				c++, *d++ = _TEXT(' ');
			while ( c < this->coltbl[COL_4] - 1 )
				c++, *d++ = _TEXT('.');
			while ( c < this->coltbl[COL_4]    )
				c++, *d++ = _TEXT(' ');

			{
				int			ch;

				s =  tbl[i].text;
				while ( (ch = *s++) != _TEXT('\0') ) 
				{
					while ( c < this->coltbl[COL_4] )
						c++, *d++ = _TEXT(' ');
					c++; *d++ = ch;
					if ( ch == _TEXT('\n') )
						c = 0;
				}
			}
			*d = _TEXT('\0');
			_putts(buffer);
		}
	}

	return;
}



#if 0
// cl getopt.cpp /D_DEBUG /DEBUG /MLd
int		main(int argc, const _TCHAR * argv[])
{
	enum	OptType
	{
		OPT_ARRAY      ,
		OPT_DEV        ,
		OPT_BLOCKSIZE  ,
		OPT_BLOCKNUMBER,
		OPT_CAPACITY   ,
		OPT_CREATEFILE ,
		OPT_CRYPTO     ,
	};
	GetOpt::AddItemParam	tbl[] =
	{
		{ OPT_ARRAY      , _TEXT("array"      ), NULL, GetOpt::VAR_TEXT  , false, { NULL,0, 0} },
		{ OPT_DEV        , _TEXT("dev"        ), NULL, GetOpt::VAR_TEXT  , false, { NULL,0, 0} },
		{ OPT_BLOCKSIZE  , _TEXT("blockSize"  ), NULL, GetOpt::VAR_INT   , true , { NULL,512,0,} },
		{ OPT_BLOCKNUMBER, _TEXT("blockNumber"), NULL, GetOpt::VAR_INT64 , false, { NULL,0,0,} },
		{ OPT_CAPACITY   , _TEXT("capacity"   ), NULL, GetOpt::VAR_INT64 , false, { NULL,0,0,} },
		{ OPT_CREATEFILE , _TEXT("createfile" ), _TEXT("c"), GetOpt::VAR_FLAG  , true , { NULL,0,0,} },
		{ OPT_CRYPTO     , _TEXT("crypto"     ), NULL, GetOpt::VAR_TEXT  , true , { _TEXT("blowfish"),0,0,} },
	};
	GetOpt	getopt;
	int		an;

	if ( getopt.AddItem(sizeof(tbl)/sizeof(*tbl), tbl) < 0 )
	{
		printf("getopt.AddItem error!\n");
		return 1;
	}

	an = 1;
	getopt.SetCurPos(an);
	while ( getopt.GetCurPos() < argc )
	{
		GetOpt::Item	*item;
//		const _TCHAR	*s;

		if ( (item = getopt.ParseArg(an, argc, argv)) == NULL )
		{
			if ( getopt.last_param != NULL )
			{	// parameter
				printf("No option Parameter \"%s\".\n", getopt.last_param );
			} else if ( getopt.err_option != NULL || getopt.err_param != NULL )
			{	// error
				if ( getopt.err_option != NULL )
					printf("Option error \"%s\".\n", getopt.err_option );
				if ( getopt.err_param != NULL )
					printf("Parameter error \"%s\".\n", getopt.err_param );
				break;
			} else
			{
				printf("End of parameter,\n");
				break;
			}
		} else
		{
			printf("Option item \"%s\".\n", item->name );
		}
	}

	return 0;
}
#endif
