//////////////////////////////////////////////////////////////////////////
//	"pglib.cpp"
//------------------------------------------------------------------------
//	$Id: pglib.cpp 103 2006-04-21 04:34:45Z taro $
//////////////////////////////////////////////////////////////////////////

#include	<stdlib.h>
#include	"pglib.h"
#if	defined(WIN32)
# include	<crtdbg.h>
#endif

#if	defined(_DEBUG) && defined(WIN32)
#	define	DBGREPORT(param_)	_CrtDbgReport param_
#else
#	define	DBGREPORT(param_)
# ifndef	_ASSERT
#	define	_ASSERT
# endif
#endif

//////////////////////////////////////////////////////////////////////////
//	PgLib

PgLib::PgLib()
{
	unsigned	n;

	this->conn = NULL;
	this->lastResult = NULL;
	this->numResult = 0;

	for ( n = 0; n < PGTYPE_HASH_MAX; n++ )
		this->pgtype.tbl[n] = NULL;
	this->pgtype.num = 0;
	this->init_pgtype();

	this->work_level = 0;
}

PgLib::~PgLib()
{
	this->clear_result();
	if ( this->conn != NULL )
		this->Finish();
	this->clear_pgtype();
}

//////////////////////////////////////////////////////////////////////////
//	ResultItem

PgLib::ResultItem::ResultItem()
{
	this->next = this->prev = NULL;
	this->pglib = NULL;
}

PgLib::ResultItem::~ResultItem()
{
	if ( this->pglib != NULL && (this->next != NULL || this->prev != NULL) )
		this->pglib->detach_result(this);
}

void	PgLib::attach_result(PgLib::ResultItem * result)
{
	result->next = NULL;
	if ( (result->prev = this->lastResult) != NULL )
		result->prev->next = result;
	this->lastResult = result;
	this->numResult++;
}

void	PgLib::detach_result(PgLib::ResultItem * result)
{
	if ( result == NULL || result->pglib != this )
		return;
	if ( result->next != NULL )
		result->next->prev = result->prev;
	if ( result->prev != NULL )
		result->prev->next = result->next;
	if ( this->lastResult == result )
		this->lastResult = result->prev;
	this->numResult--;
	delete result;

	return;
}

PgLib::ResultItem *	PgLib::create_result(libpq::PGresult * pg_result)
{
	PgLib::ResultItem *	result;

	if ( (result = new PgLib::ResultItem) == NULL )
		return NULL;
	result->pglib = this;
	(PgLib::Result &)*result = pg_result;
	this->attach_result(result);

	return result;
}

int		PgLib::destroy_result(PgLib::ResultItem * result)
{
	if ( result == NULL || result->pglib != this )
		return -1;
	this->detach_result(result);
	delete result;

	return 0;
}

void		PgLib::clear_result(void)
{
	PgLib::ResultItem	*result, *result_prev;

	result = this->lastResult;
	while ( result != NULL )
	{
		result_prev = result->prev;
		this->detach_result(result);
		delete result;
		result = result_prev;
	}
	this->lastResult = NULL;
	this->numResult = 0;
}

//////////////////////////////////////////////////////////////////////////
//	Connection

int		PgLib::Connectdb(const char * info)
{
	if ( this->conn != NULL )
		return -1;
	if ( (this->conn = libpq::PQconnectdb(info)) == NULL )
		return -1;
	libpq::PQsetNoticeReceiver(this->conn, NoticeReceiver, this);
	libpq::PQsetNoticeProcessor(this->conn, NoticeProcessor, this);
	return 0;
}

int		PgLib::Finish(void)
{
	if ( this->conn == NULL )
		return 0;
	libpq::PQfinish(this->conn);
	this->conn = NULL;
	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	Query

int			PgLib::Exec(PgLib::Result & result, const char * query)
{
	if ( this->conn == NULL )
		return -1;

#if	1
	if ( this->ExecStart() == false )
		return -1;
	if ( libpq::PQsendQuery(this->conn, query) == false )
		return -1;
	if ( (result = this->ExecFinish()) == NULL )
		return -1;
#else
	if ( (result = libpq::PQexec(this->conn, query)) == NULL )
		return -1;
#endif

	return 0;
}

PgLib::Result *	PgLib::Exec(const char * query)
{
	libpq::PGresult *	pg_result;

	if ( this->conn == NULL )
		return NULL;
	if ( this->ExecStart() == false )
		return NULL;
	if ( libpq::PQsendQuery(this->conn, query) == false )
		return NULL;
	if ( (pg_result = this->ExecFinish()) == NULL )
		return NULL;
	return this->create_result(pg_result);
}

int			PgLib::GetResult(PgLib::Result & result)
{
	if ( this->conn == NULL )
		return -1;
	if ( (result = libpq::PQgetResult(this->conn)) == NULL )
		return -1;

	return 0;
}

PgLib::Result *		PgLib::GetResult(void)
{
	libpq::PGresult *	pg_result;

	if ( this->conn == NULL )
		return NULL;
	if ( (pg_result = libpq::PQgetResult(this->conn)) == NULL )
		return NULL;
	return this->create_result(pg_result);
}

void			PgLib::ClearResult(PgLib::Result * result)
{
	if ( result == NULL )
	{	// clear all
		this->clear_result();
	} else
	{
		this->destroy_result( (PgLib::ResultItem *)result );
	}
	return;
}

int		PgLib::SendQuery(const char * query)
{
	if ( this->conn == NULL )
		return -1;

	if ( libpq::PQsendQuery(this->conn, query) == false )
		return -1;
	return 0;
}


#define PG_PROTOCOL_MAJOR(v)	((v) >> 16)
#define PG_PROTOCOL_MINOR(v)	((v) & 0x0000ffff)
#define PG_PROTOCOL(m,n)	(((m) << 16) | (n))

//
// Common code for PQexec and sibling routines: prepare to send command
//
bool	PgLib::ExecStart(void)
{
	libpq::PGresult *	pg_result;

	if ( this->conn == NULL )
		return false;

	// Silently discard any prior query result that application didn't
	// eat. This is probably poor design, but it's here for backward
	// compatibility.
	while ((pg_result = libpq::PQgetResult(this->conn)) != NULL)
	{
		libpq::ExecStatusType	resultStatus;

		resultStatus = libpq::PQresultStatus(pg_result);
		libpq::PQclear(pg_result);		// only need its status
		if ( resultStatus == libpq::PGRES_COPY_IN )
		{
			if ( PG_PROTOCOL_MAJOR(libpq::PQprotocolVersion(this->conn)) >= 3 )
			{
				// In protocol 3, we can get out of a COPY IN state
				if ( libpq::PQputCopyEnd( this->conn, "COPY terminated by new PQexec") < 0 )
					return false;
				// keep waiting to swallow the copy's failure message
			} else
			{
				// In older protocols we have to punt
			//	printfPQExpBuffer(&conn->errorMessage,
			//					  libpq_gettext("COPY IN state must be terminated first\n"));
				return false;
			}
		} else if ( resultStatus == libpq::PGRES_COPY_OUT )
		{
			if ( PG_PROTOCOL_MAJOR(libpq::PQprotocolVersion(this->conn)) >= 3 )
			{
				// In protocol 3, we can get out of a COPY OUT state: we
				// just switch back to BUSY and allow the remaining COPY
				// data to be dropped on the floor.
			//	conn->asyncStatus = libpq::PGASYNC_BUSY;
				// keep waiting to swallow the copy's completion message
			}
			else
			{
				/* In older protocols we have to punt */
			//	printfPQExpBuffer(&conn->errorMessage,
			//					  libpq_gettext("COPY OUT state must be terminated first\n"));
				return false;
			}
		}
		// check for loss of connection, too
		if ( libpq::PQstatus(this->conn) == libpq::CONNECTION_BAD)
			return false;
	}

	// OK to send a command
	return true;
}

//
// Common code for PQexec and sibling routines: wait for command result
//
libpq::PGresult *	PgLib::ExecFinish(void)
{
	libpq::PGresult *	pg_result;
	libpq::PGresult *	pg_lastResult;

	// For backwards compatibility, return the last result if there are
	// more than one --- but merge error messages if we get more than one
	// error result.
	//
	// We have to stop if we see copy in/out, however. We will resume parsing
	// after application performs the data transfer.
	//
	// Also stop if the connection is lost (else we'll loop infinitely).
	pg_lastResult = NULL;
	while ((pg_result = libpq::PQgetResult(this->conn)) != NULL)
	{
		if ( pg_lastResult != NULL )
		{
			if (libpq::PQresultStatus(pg_lastResult) == libpq::PGRES_FATAL_ERROR &&
			    libpq::PQresultStatus(pg_result    ) == libpq::PGRES_FATAL_ERROR)
			{
			//	pqCatenateResultError(lastResult, result->errMsg);
				libpq::PQclear(pg_result);
				pg_result = pg_lastResult;

				// Make sure PQerrorMessage agrees with concatenated
				// result
			//	resetPQExpBuffer(&conn->errorMessage);
			//	appendPQExpBufferStr(&conn->errorMessage, result->errMsg);
			} else
				libpq::PQclear(pg_lastResult);
		}
		pg_lastResult = pg_result;
		if ( libpq::PQresultStatus(pg_result) == libpq::PGRES_COPY_IN ||
			 libpq::PQresultStatus(pg_result) == libpq::PGRES_COPY_OUT ||
			 libpq::PQstatus(conn) == libpq::CONNECTION_BAD)
			break;
	}

	return pg_lastResult;
}

//////////////////////////////////////////////////////////////////////////
//	Library

int		PgLib::BeginWork(void)
{
	if ( this->conn == NULL )
		return -1;
	if ( this->work_level == 0 )
	{
		PgLib::Result	result;

		this->work_rollback = false;
		if ( this->Exec(result, "BEGIN") < 0
		  || result.GetStatus() != PgLib::PGRES_COMMAND_OK )
			return -1;
		result.Clear();
	}
	this->work_level++;
	return 0;
}

int		PgLib::CommitWork(void)
{
	if ( this->conn == NULL )
		return -1;
	if ( this->work_level > 0 )
	{
		this->work_level--;
		if ( this->work_level == 0 )
		{
			PgLib::Result	result;

			if ( this->work_rollback == false )
			{
				if ( this->Exec(result, "COMMIT") < 0 )
					return -1;
			} else
			{
				if ( this->Exec(result,"ROLLBACK") < 0 )
					return -1;
			}
			if ( result.GetStatus() != PgLib::PGRES_COMMAND_OK )
				return -1;
			result.Clear();
		}
	}
	return 0;
}

int		PgLib::RollbackWork(void)
{
	if ( this->conn == NULL )
		return -1;
	if ( this->work_level > 0 )
	{
		this->work_level--;
		if ( this->work_level == 0  )
		{
			PgLib::Result	result;

			if ( this->Exec(result, "ROLLBACK") < 0
			  || result.GetStatus() != PgLib::PGRES_COMMAND_OK )
				return -1;
		} else
			this->work_rollback = true;
	}
	return 0;
}

void	PgLib::NoticeReceiver(void * arg, const libpq::PGresult *res)
{
	if ( arg != NULL )
		((PgLib *)arg)->NoticeReceiver(res);
	return;
}

void PgLib::NoticeProcessor(void * arg, const char *message)
{
	if ( arg != NULL )
		((PgLib *)arg)->NoticeProcessor(message);
	return;
}

void	PgLib::NoticeReceiver(const libpq::PGresult *res)
{
	return;
}

void	PgLib::NoticeProcessor(const char * message)
{
	if ( message != NULL )
	{
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"PgLib::NoticeProcessor: %s\n", message));
	}
	return;
}

