//////////////////////////////////////////////////////////////////////////
//	"smtplib.h"
//------------------------------------------------------------------------
//	$id: $
//////////////////////////////////////////////////////////////////////////
#ifndef	SMTPLIB_H_INCLUDED
#define	SMTPLIB_H_INCLUDED

#include	"socklib.h"
#include	"varbuf.h"
#include	<time.h>

//#define	SMTPLIB_LOG_SEND
#define	SMTPLIB_VARBUF_BASE64

class	SmtpLib : public SockLib
{
public:
	enum poplib_enum
	{
		TIMEOUT         = 30000,
		LOOP_TIMEOUT    =   100,
		SYSWAIT         =    50,
		SENDWAIT        =     0,
		PORT_SMTP       =    25,
		BOUNDARY_MAX    =   256,
		BASE64_MAXLEN   =    75,
		QP_MAXLEN       =    75,
		TICK_PER_SECOND =  1000,
		CONNECT_RETRY   =     3,
		CONNECT_WAIT    =  3000,
	};
	typedef unsigned	TickTime;
	typedef int (SmtpLib::* LineCheckFunc)(unsigned, const void *);

	enum encode_enum
	{
		ENCODE_OFF   ,
		ENCODE_7BIT  ,
		ENCODE_8BIT  ,
		ENCODE_QP    ,	// Quoted Printable
		ENCODE_BASE64,
	};
	struct PartData
	{
		const TCHAR *	content_type;
		const TCHAR *	content_name;
		const TCHAR *	content_id;
		unsigned		encode;
		const TCHAR		*filename;
		unsigned		datasize;
		const void *	data;
		PartData()
		{
			this->content_type = this->content_name = NULL;
			this->content_name = NULL;
			this->content_id   = NULL;
			this->encode       = ENCODE_7BIT;
			this->filename     = NULL;
			this->datasize     = 0;
			this->data         = NULL;
		}
	};

public:
#if	! defined(SMTPLIB_VARBUF_BASE64)
	static const unsigned	base64tbl[64+1];
#endif
	static const char *		MimeStart;
	static const char *		MimeEnd;
	static const char *		EscIn;
	static const char *		EscKana;
	static const char *		EscOut;
	static const char *		SO;
	static const char *		SI;
	static const unsigned	MimeStartSize;
	static const unsigned	MimeEndSize;
	static const unsigned	EscInSize;
	static const unsigned	EscKanaSize;
	static const unsigned	EscOutSize;
	static const unsigned	SOSize;
	static const unsigned	SISize;
#if	! defined(SMTPLIB_VARBUF_BASE64)
	static inline unsigned 	GetBase64Char(unsigned b) { return base64tbl[b & 63]; }
#endif

public:
	SockAddr			smtp_addr;
	int					timeout;
	unsigned			sendwait;
	VarBuffer			vbuf;

	int					rescode;

#if	defined(WIN32)
	HANDLE				event;
#endif

public:
	SmtpLib();
	virtual ~SmtpLib();

#if	defined(WIN32)
	void		SetEventHandle(HANDLE handle) { this->event = handle; }
#endif
	int			CheckAbort(void);
	int 		Create(void);
	int			Connect(void);
	int			CheckConnect(void);
	int			SendMail(const TCHAR * from, const TCHAR * to, const TCHAR * text, const TCHAR * header, BOOL connect = TRUE);
	int			Quit(unsigned msec);
	int			Wait(unsigned msec);
	int			CreateMutipartMessage(VarBuffer & header, VarBuffer & body, const char * text, unsigned partnum, PartData * parttbl);
	static int	GetBoundary(char * buffer, unsigned id);

private:
	int			CheckRecv(unsigned msec);
	int			CheckRecv(unsigned msec, LineCheckFunc func);
	int			linecheck_func(unsigned len, const char * text);
	int			CheckSend(unsigned size, const TCHAR * buffer, unsigned msec);
	int			SendText(unsigned size, const TCHAR * buffer, unsigned msec);

public:
	static int	EncodeHeaderText(const TCHAR * name, VarBuffer & wbuf);
	static int	IsSpecialChar(int ch);
	static int	MimeEncode(const TCHAR * src, VarBuffer & wbuf, BOOL dqenc = FALSE);
	static int	Base64Encode(unsigned size, const void * buffer, VarBuffer & wbuf, unsigned maxlen = 0)
#if	defined(SMTPLIB_VARBUF_BASE64)
	{ return VarBuffer::Base64Encode(size, buffer, wbuf,  maxlen); }
#endif
	;
	static int	QuotedPrintableEncode(unsigned size, const void * buffer, VarBuffer & wbuf, unsigned maxlen = 0);
	static int	Sjis2JisText(const TCHAR * src, VarBuffer & wbuf);
	static int	Sjis2Jis(int ch);
	static int	Jis2Sjis(int ch);
	static TickTime	GetTick(void);
	static void		Sleep(unsigned msec);
//	static const TCHAR *	GetDayTime(const time_t * tm, VarBuffer & vbuf);
};


#endif
