//////////////////////////////////////////////////////////////////////////
//	"sql_datavalue.cpp"
//------------------------------------------------------------------------
//	$Id: sql_datavalue.cpp 11 2006-03-03 14:20:05Z taro $
//////////////////////////////////////////////////////////////////////////

#include	"sqllib.h"
#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#if	defined(WIN32)
# include	<tchar.h>
#endif

#if 1 && defined(_DEBUG)
# if	defined(WIN32)
#  include	<crtdbg.h>
#  define	DBGREPORT(param_)	_CrtDbgReport param_
# endif
#else
# define	DBGREPORT(param_)
#endif

//////////////////////////////////////////////////////////////////////////
//	Srep

SqlLib::DataValue::Srep::Srep()
{
	this->dtype = DTYPE_NONE;
	this->size  = 0;
	this->ref = 1;

	return;
}

SqlLib::DataValue::Srep::~Srep()
{
	this->Clear();
	return;
}

void	SqlLib::DataValue::Srep::Clear(void)
{
	switch ( this->dtype )
	{
	case DTYPE_CHAR:
	case DTYPE_TEXT:
	case DTYPE_VARCHAR:
		if ( this->v.text != NULL )
		{
			delete [] this->v.text;
			this->v.text = NULL;
		}
		break;

	case DTYPE_BINARY:
	case DTYPE_VARBINARY:
		if ( this->v.p != NULL )
		{
			delete [] (BYTE *)this->v.p;
			this->v.p = NULL;
		}
		break;

	default:
		break;
	}
	this->dtype = DTYPE_NONE;
	return;
}

int		SqlLib::DataValue::Srep::SetTextBuffer(DataType dtype, const TCHAR *s, int size)
{
	size_t	len;

	this->Clear();
	if ( size < 0 )
	{
		if ( s != NULL )
			len = _tcslen(s);
		else
			len = 0;
	} else
		len = size;
	this->dtype = dtype;
	if ( (this->v.text = new TCHAR [len + 1]) == NULL )
		return -1;
	this->v.text[len] = TEXT('\0');
	if ( s != NULL )
		_tcsncpy(this->v.text, s, len);
	else
		this->v.text[0] = TEXT('\0');
	this->size = size;

	return 0;
}

int		SqlLib::DataValue::Srep::SetBinaryBuffer(DataType dtype, const void * p, int size)
{
	this->Clear();
	this->dtype = dtype;
	if ( (this->v.p = new BYTE [size]) == NULL )
		return -1;
	memcpy(this->v.p, p, size);
	this->size = size;

	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	DataValue

SqlLib::DataValue::DataValue()
{
	this->p = NULL;
	return;
}

SqlLib::DataValue::~DataValue()
{
	if ( this->p != NULL )
	{
		if ( --(this->p->ref) == 0 )
		{
			delete this->p;
			this->p = NULL;
		}
	}
	return;
}

SqlLib::DataValue & 		SqlLib::DataValue::operator = (const SqlLib::DataValue & src)
{
	if ( this->p == src.p )
		return *this;
	if ( ((DataValue *)(&src))->check_srep() < 0 )
		return *this;
	++(src.p->ref);
	if ( this->p != NULL )
	{
		if ( --(this->p->ref) == 0 )
		{
			delete this->p;
			this->p = NULL;
		}
	}
	this->p = src.p;
	return *this;
}

void		SqlLib::DataValue::Clear(void)
{
	if ( this->p != NULL )
		this->p->Clear();
	return;
}


#if 0
SqlLib::DataValue & 		SqlLib::DataValue::operator = (const SqlLib::DataValue & src)
{
	this->Clear();
	switch ( src.dtype )
	{
	case DTYPE_CHAR:
	case DTYPE_TEXT:
	case DTYPE_VARCHAR:
		this->SetTextBuffer(src.dtype, src.v.text, src.size);
		break;

	case DTYPE_BINARY:
	case DTYPE_VARBINARY:
		this->SetBinaryBuffer(src.dtype, src.v.p, src.size);
		break;

	default:
		this->dtype = src.dtype;
		this->v     = src.v;
		this->size  = src.size;
		break;
	}
	return *this;
}

void	SqlLib::DataValue::SetTextBuffer(DataType dtype, const TCHAR *s, int size)
{
	size_t	len;

	this->Clear();
	if ( size < 0 )
	{
		if ( s != NULL )
			len = _tcslen(s);
		else
			len = 0;
	} else
		len = size;
	this->dtype = dtype;
	if ( (this->buffer = new TCHAR [len + 1]) != NULL )
	{
		((TCHAR *)(this->buffer))[len] = TEXT('\0');
		if ( s != NULL )
			_tcsncpy((TCHAR *)this->buffer, s, len);
		else
			*(TCHAR *)(this->buffer) = TEXT('\0');
		this->v.text = (TCHAR *)this->buffer;
	} else
	{
		this->v.text = s;
	}
	this->size = size;
}

void	SqlLib::DataValue::SetBinaryBuffer(DataType dtype, const void * p, int size)
{
	this->Clear();
	this->dtype = dtype;
	if ( (this->buffer = new TCHAR [size]) != NULL )
	{
		memcpy(this->buffer, p, size);
		this->v.p = this->buffer;
	} else
	{
		this->v.p = p;
	}
	this->size = size;
}

#endif
