//////////////////////////////////////////////////////////////////////////
//	"sqstr.cpp" -- string class
//------------------------------------------------------------------------
//	$Id: sqstr.cpp 21 2006-03-04 08:46:09Z taro $
//////////////////////////////////////////////////////////////////////////

#include	"sqstr.h"
#include	<stdlib.h>
#include	<string.h>
#if	defined(WIN32)
# include	<windows.h>
#elif	defined(LINUX)
# include	"../linux/defs.h"
# include	<limits.h>
#endif

SQStr::SQStr()
{
	this->p = NULL;
	this->error = SQERR_NOERR;
}

SQStr::SQStr(const SQStr & str)
{
	if ( str.p != NULL )
		((SQStr &)str).p = new srep;
	++(str.p->ref);
	this->p = str.p;
	this->error = this->p != NULL ? SQERR_NOERR : SQERR_OUT_OF_MEMORY;
}

SQStr::SQStr(const TCHAR * s)
{
	this->p = new srep;
	if ( this->p == NULL )
	{
		this->error = SQERR_OUT_OF_MEMORY;
		return;
	}
	this->SetText(s);
}

SQStr::~SQStr()
{
	if ( this->p != NULL )
	{
		if ( --(this->p->ref) == 0 )
		{
			this->ClearText();
			delete this->p;
			this->p = NULL;
		}
	}
}

SQStr &		SQStr::operator = (const SQStr & str)
{
	if ( str.p == NULL )
		((SQStr &)str).p = new srep;
	++(str.p->ref);
	if ( this->p != NULL )
	{
		if ( --(this->p->ref) == 0 )
		{
			this->ClearText();
			delete this->p;
			this->p = NULL;
		}
	}
	this->p = str.p;
	return *this;
}

SQStr &		SQStr::operator = (const TCHAR * s)
{
	if ( this->p != NULL )
	{
		if ( this->p->ref > 1 )
		{
			--(this->p->ref);
			this->p = new srep;
			if ( this->p == NULL )
			{
				this->error = SQERR_OUT_OF_MEMORY;
				return *this;
			}
		} else
			this->ClearText();
	}
	this->SetText(s);
	return *this;
}

void	SQStr::ClearText(void)
{
	if ( this->p != NULL )
	{
		if ( this->p->buffer != NULL )
		{
			if ( (this->p->mode & MODE_FLAG_ALLOC) != 0 )
				delete [] (TCHAR *)this->p->buffer;
			this->p->buffer = NULL;
		}
		this->p->length = 0;
	}
}

TCHAR *	SQStr::SetText(const TCHAR * s)
{
	this->ClearText();
	if ( s != NULL )
	{
		if ( this->p == NULL )
			this->p = new srep;
		this->p->length = (unsigned)::_tcslen(s);
		if ( (this->p->buffer = new TCHAR[this->p->length+1]) == NULL )
		{
			this->p->length = 0;
			this->error = SQERR_OUT_OF_MEMORY;
			return NULL;	// out of memory
		}
		this->p->mode |= MODE_FLAG_ALLOC;
		::_tcscpy( (TCHAR *)this->p->buffer, s );
	}
	this->error = SQERR_NOERR;
	return this->p != NULL ? (TCHAR *)(this->p->buffer) : NULL;
}

SQStr &	SQStr::SetText(const SQStr & s)
{
	unsigned	size;

	if ( this->p == s.p )
		return *this;
	if ( s.p != NULL )
	{
		switch ( s.p->chcode )
		{
			case SQCHCODE_UNICODE:
				size = (s.p->length + 1) * sizeof(wchar_t);
				break;
			case SQCHCODE_ANSI:
				size = (s.p->length + 1) * sizeof(char);
				break;
			default:
				this->ClearText();
				this->error = SQERR_CHCODE_MISMATCH;
				return *this;
		}
		if ( this->TransferToBuffer(size, s.p->buffer, s.p->length, s.p->chcode) < 0 )
			this->ClearText();
	} else
		this->ClearText();
	return *this;
}

unsigned	SQStr::GetBufferSize(void)
{
	if ( this->p == NULL || this->p->buffer == NULL )	// BugFix 2002.05.31
		return 0;
	switch ( this->p->chcode )
	{
		case SQCHCODE_UNICODE:
			return (this->p->length + 1) * sizeof(wchar_t);
		case SQCHCODE_ANSI:
			return (this->p->length + 1) * sizeof(char);
	}
	return 0;
}

void *	SQStr::SetBuffer(const void * buffer, unsigned length, int chcode)
{
	if ( this->p == NULL )
	{
		if ( (this->p = new srep) == NULL )
			return NULL;
	} else
		this->ClearText();
	this->p->buffer = (void *)buffer;
	this->p->mode  &= ~MODE_FLAG_ALLOC;
	this->p->length = length;
	this->p->chcode = chcode;
	return this->p->buffer;
}

void *	SQStr::CreateBuffer(unsigned size, unsigned length, int chcode)
{
	if ( this->p != NULL )
		this->ClearText();
	else
		this->p = new srep;
	if ( (this->p->buffer = new BYTE [size]) == NULL )
	{
		this->error = SQERR_OUT_OF_MEMORY;
		return NULL;
	}
	this->p->mode |= MODE_FLAG_ALLOC;
	this->p->length = length;
	this->p->chcode = chcode;
	return this->p->buffer;
}

int		SQStr::TransferToBuffer(unsigned byte, const void * buffer, int length, int chcode)
{
	if ( this->p != NULL )
		this->ClearText();
	else
		this->p = new srep;
	if ( buffer != NULL )
	{
		this->p->length = length;
		if ( (this->p->buffer = new BYTE [byte]) == NULL )
		{
			this->p->length = 0;
			this->error = SQERR_OUT_OF_MEMORY;
			return -1;	// out of memory
		}
		this->p->mode |= MODE_FLAG_ALLOC;
		::memcpy( p->buffer, buffer, byte);
		this->p->chcode = chcode;
	}
	this->error = SQERR_NOERR;
	return 0;
}


#if	defined(LINUX)
#ifndef	CP_OEMCP
# define	CP_OEMCP	(0)
#endif
static int	WideCharToMultiByte(unsigned codepage, DWORD flags, const wchar_t * wbuf, int wlen, char * abuf, int abyte, const char *defchar, BOOL * useDefChar)
{
	int				wch, wpos, alen = 0;

	if ( wbuf == NULL )
		return 0;	// error
	for ( wpos = 0; wpos < wlen; wpos++ )
	{
		if ( abuf != NULL && alen >= abyte )
			break;
		wch = wbuf[wpos];
		if ( wch < 0x100 )
		{
			if ( abuf != NULL && alen < abyte )
				abuf[alen] = (char)wch;
			alen++;
		} else
		{
			if ( abuf != NULL && alen + 1 < abyte )
			{
				abuf[alen  ] = '@';
				abuf[alen+1] = '@';
			}
			alen += 2;
		}
	}
	return alen;
}

static int	MultiByteToWideChar(unsigned codepage, DWORD flags, const char *abuf, int abyte, wchar_t *wbuf, int wlen)
{
	int			ach, apos, wret = 0;

	if ( abuf == NULL )
		return 0;	// error
	apos = 0;
	while ( apos < abyte )
	{
		if ( wbuf != NULL && wret >= wlen )
			break;
		ach = abuf[apos++];
		if ( apos < abyte && _ismbblead(ach) && _ismbbtrail(abuf[apos]) )
		{	// {
			if ( wbuf != NULL )
				wbuf[wret] = (ach<<8) + abuf[apos];
			wret++;
			apos++;
		} else
		{
			if ( wbuf != NULL )
				wbuf[wret] = ach;
			wret++;
		}
	}
	return wret;
}
#endif

SQStr &		SQStr::ChangeChcode(const SQStr & text, int chcode)
{
	if ( text.GetLength() == 0)
	{
		this->SetText(text);
		this->SetChcode(chcode);
		return *this;
	}

	switch ( chcode )
	{
		case SQCHCODE_UNICODE:
			switch ( text.GetChcode() )
			{
				case SQCHCODE_UNICODE:
					return this->SetText(text);

				case SQCHCODE_ANSI:
				{
					wchar_t	   *wbuf;
					unsigned	wlen, size;
					int			len;

					wlen = text.GetLength() + 1;
					if ( (wbuf = new wchar_t [wlen]) == NULL )
					{
						this->error = SQERR_OUT_OF_MEMORY;
						goto ERR;
					}
					len = ::MultiByteToWideChar(CP_OEMCP, 0, (char *)text.GetBuffer(), text.GetLength(), (wchar_t *)wbuf, wlen);
					if ( len > 0 )
					{
						if ( len < (int)wlen )
						{
							wbuf[len] = 0;
							size = (len+1) * sizeof(wchar_t);
						} else
							size = wlen * sizeof(wchar_t);
						if ( this->TransferToBuffer( size, wbuf, len, SQCHCODE_UNICODE) < 0 )
							goto ERR;
					}
					delete [] wbuf;
					break;
				}

				default:
					this->error = SQERR_CHCODE_MISMATCH;
					goto ERR;
			}
			break;

		case SQCHCODE_ANSI:
			switch ( text.GetChcode() )
			{
				case SQCHCODE_UNICODE:
				{
					char	   *abuf;
					unsigned	alen, size;
					int			len;

					alen = (text.GetLength() + 1) * 2;
					if ( (abuf = new char [alen]) == NULL )
					{
						this->error = SQERR_OUT_OF_MEMORY;
						goto ERR;
					}
					len = ::WideCharToMultiByte(CP_OEMCP, 0, (wchar_t *)text.GetBuffer(), text.GetLength(), abuf, alen, NULL, NULL);
					if ( len > 0 )
					{
						if ( len < (int)alen )
						{
							abuf[len] = 0;
							size = (len+1) * sizeof(char);
						} else
							size = alen * sizeof(char);
						if ( this->TransferToBuffer( size, abuf, len, SQCHCODE_ANSI) < 0 )
							goto ERR;
					}
					delete [] abuf;
					break;
				}

				case SQCHCODE_ANSI:
					return this->SetText(text);

				default:
					this->error = SQERR_CHCODE_MISMATCH;
					goto ERR;
			}
			break;

		default:
			this->error = SQERR_CHCODE_MISMATCH;
			goto ERR;
	}
	return *this;

ERR:
	this->ClearText();
	return *this;
}

SQStr &		SQStr::ChangeChcode(int chcode)
{
	if ( this->GetLength() == 0 )
	{
		this->SetChcode(chcode);
		return *this;
	}
	return this->ChangeChcode(*this, chcode);
}


int		SQStr::GetPtrChar(void * ptr) const
{
	if ( ptr == NULL || this->p == NULL || ptr < this->p->buffer )
		return -1;

	switch ( this->p->chcode )
	{
		case SQCHCODE_UNICODE:
			if ( (wchar_t *)ptr >= (wchar_t *)(this->p->buffer) + this->p->length )
				return -1;
			return *(wchar_t *) ptr;

		case SQCHCODE_ANSI:
		{
			unsigned char	*s = (unsigned char *)ptr;

			if ( (char *)ptr >= (char *)(this->p->buffer) + this->p->length )
				return -1;
			if ( _ismbblead(s[0]) && _ismbbtrail(s[1]) )
				return  (s[0]<<8) | s[1];
			else
				return s[0];
			break;
		}
	}
	return -1;	// error
}

void *	SQStr::GetPtrNext(const void * ptr) const
{
	if ( ptr == NULL || this->p == NULL || ptr < this->p->buffer )
		return NULL;
	switch ( this->p->chcode )
	{
		case SQCHCODE_UNICODE:
		{
			wchar_t	   *s = (wchar_t *) ptr;

			if ( s >= (wchar_t *)(this->p->buffer) + this->p->length )
				return NULL;
			++s;
			ptr = s;
			break;
		}

		case SQCHCODE_ANSI:
		{
			unsigned char	*s = (unsigned char *)ptr;

			if ( (char *)ptr >= (char *)(this->p->buffer) + this->p->length )
				return NULL;
			if ( _ismbblead(s[0]) && _ismbbtrail(s[1]) )
				++s;
			++s;
			ptr = s;
			break;
		}

		default:
			return NULL;
	}
	return (void *)ptr;
}

void *	SQStr::PutPtrChar(void * ptr, int ch, unsigned * psize, unsigned * plen)
{
	int		size = 0, len = 0;

	if ( ptr == NULL )
		return NULL;

	switch ( this->GetChcode() )
	{
		case SQCHCODE_UNICODE:
		{
			wchar_t	   *s = (wchar_t *) ptr;

			*s++ = ch;
			ptr = s;
			size = sizeof(wchar_t);
			len = 1;
			break;
		}

		case SQCHCODE_ANSI:
		{
			unsigned char	*s = (unsigned char *)ptr;

			if ( ch >= 256 )
			{
				*s++ = (ch>>8);
				*s++ = (ch & 0xFF);
				size = len = 2;
			} else
			{
				*s++ = ch;
				size = len = 1;
			}
			ptr = s;
			break;
		}

		default:
			return NULL;
	}
	if ( plen != NULL )
		*plen += len;
	if ( psize )
		*psize += size;
	return ptr;
}

int		SQStr::StrCmp(const TCHAR *s1, const TCHAR * s2, unsigned n)
{
	int			ret;
	unsigned	len;

	if ( s1 == NULL && s2 != NULL )
		return -1;
	if ( s1 != NULL && s2 == NULL )
		return 1;
	if ( s1 == s2 )
		return 0;
	len = 0;
	ret = 0;
#if	SQCHCODE_DEFAULT==SQCHCODE_SJIS
	{
		int		ch1, ch2;

		while ( len < n )
		{
			ch1 = *(unsigned char *)s1; s1++; len++;
			if ( _ismbblead(ch1) && _ismbbtrail(*(unsigned char*)s1) )
			{
				ch1 = (ch1<<8) | (*(unsigned char *)s1);
				s1++; len++;
			}

			ch2 = *(unsigned char *)s2; s2++;
			if ( _ismbblead(ch2) && _ismbbtrail(*(unsigned char*)s2) )
			{
				ch2 = (ch2<<8) | (*(unsigned char *)s2);
				s2++;
			}
			if ( (ret = ch1 - ch2) != 0 )
				break;
			if ( ch1 == 0 || ch2 == 0 )
				break;
		}
	}
#endif

#if SQCHCODE_DEFAULT==SQCHCODE_UNICODE
#error
#endif

	return ret;
}

int		SQStr::StrCmp(const TCHAR *s1, const TCHAR * s2)
{
	return SQStr::StrCmp(s1, s2, UINT_MAX);
}

int		SQStr::StrCmpi(const TCHAR *s1, const TCHAR * s2, unsigned n)
{
	int			ret;
	unsigned	len;

	if ( s1 == NULL && s2 != NULL )
		return -1;
	if ( s1 != NULL && s2 == NULL )
		return 1;
	if ( s1 == s2 )
		return 0;
	len = 0;
	ret = 0;
#if	SQCHCODE_DEFAULT==SQCHCODE_SJIS
	{
		int		ch1, ch2;

		while ( len < n )
		{
			ch1 = *(unsigned char *)s1; s1++; len++;
			if ( _ismbblead(ch1) && _ismbbtrail(*(unsigned char*)s1) )
			{
				ch1 = (ch1<<8) | (*(unsigned char *)s1);
				s1++; len++;
			} else if ( ch1 >= 'a' && ch1 <= 'z' )
				ch1 = 'A' + (ch1 - 'a');

			ch2 = *(unsigned char *)s2; s2++;
			if ( _ismbblead(ch2) && _ismbbtrail(*(unsigned char*)s2) )
			{
				ch2 = (ch2<<8) | (*(unsigned char *)s2);
				s2++;
			} else if ( ch2 >= 'a' && ch2 <= 'z' )
				ch2 = 'A' + (ch2 - 'a');
			if ( (ret = ch1 - ch2) != 0 )
				break;
			if ( ch1 == 0 || ch2 == 0 )
				break;
		}
	}
#endif

#if SQCHCODE_DEFAULT==SQCHCODE_UNICODE
#error
#endif

	return ret;
}

int		SQStr::StrCmpi(const TCHAR *s1, const TCHAR * s2)
{
	return SQStr::StrCmpi(s1, s2, UINT_MAX);
}

