//////////////////////////////////////////////////////////////////////////
//	"devctl_blkdev.cpp"
//------------------------------------------------------------------------
//	$Id: devctl_blkdev.cpp 86 2006-03-30 09:13:39Z taro $
//////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#	include "config.h"
#	undef	HAVE_CONFIG_H
#endif
#include	"devctl.h"
#include	"linux/hdreg.h"
#include	"linux/cdrom.h"

//////////////////////////////////////////////////////////////////////////
//	Block Device

int		DevCtl::blkdev_opendevice(void)
{
	struct hd_driveid	drvid;
	off_t				fpos;

	if ( ! S_ISBLK(this->statbuf.st_mode) )
		return -1;
	if ( this->cdrom_opendevice() == 0 )
		return 0;	// CD-ROM
	this->htype = HT_BLKDEV;
	this->blk.blkLen = 512;
	this->blk.blkNum = 0;
	if ( (fpos = lseek(this->fd, (off_t)0, SEEK_END)) < 0 )
		fpos = 0;
	lseek(this->fd, (off_t)0, SEEK_SET);
#if	defined(DEBUG)
	fprintf(stderr, "DEBUG: lseek fpos %lld\n", (Int64)fpos);
#endif
	memset( &drvid, 0, sizeof(drvid));
	if ( ioctl(this->fd, HDIO_GET_IDENTITY, &drvid) == 0 )
	{
		if ( drvid.sector_bytes > 0 )
			this->blk.blkLen = drvid.sector_bytes;
		this->blk.blkNum = (Int64)fpos / (Int64) this->blk.blkLen;
#if	defined(DEBUG)
		fprintf(stderr, "DEBUG: sector_byte %u fpos/sector_byte %lld\n"
		  , (unsigned)this->blk.blkLen
		  , (Int64)this->blk.blkNum);
#endif
#if	defined(HAVE_HD_DRIVEID_LBA_CAPACITY_2)
		if ( drvid.lba_capacity_2 > this->blk.blkNum )
			this->blk.blkNum = drvid.lba_capacity_2;
#else
		if ( *(Int64 *) &((WORD *)&drvid)[100] > this->blk.blkNum )
			this->blk.blkNum = *(Int64 *) &((WORD *)&drvid)[100];
#endif
		else if ( drvid.lba_capacity > this->blk.blkNum )
			this->blk.blkNum = drvid.lba_capacity;
	} else
		this->blk.blkNum = (Int64)fpos / (Int64)this->blk.blkLen;
	return 0;
}

int		DevCtl::SetDefaultBlockSize(unsigned size)
{
	if ( this->fd >= 0 && this->htype != HT_BLKDEV )
		return -1;
//	this->blk.blkNum = this->blk.blkNum * this->blk.blkLen / size;
	this->blk.blkLen = size;
	return 0;
}

int		DevCtl::BlkDevGetCapacity(unsigned & blkLen, Int64 & blkNum)
{
	if ( this->fd < 0 || this->htype != HT_BLKDEV )
		return -1;
	blkLen = this->blk.blkLen;
	blkNum = this->blk.blkNum;
	return 0;
}

int		DevCtl::BlkDeviInquiry(DevCtl::SCSIINQUIRY & inq)
{
	struct hd_driveid	drvid;
	char				tmp[256];

	if ( this->fd < 0 || S_ISBLK(this->statbuf.st_mode) == FALSE )
		return -1;
	memset( &drvid, 0, sizeof(drvid));
	if ( ioctl(this->fd, HDIO_GET_IDENTITY, &drvid) < 0 )
	{
		return 0;
	} else
	{
		memset(&inq, 0, sizeof(inq));
		sprintf(tmp, "%.40s %.8s", (char *)drvid.model, (char *)drvid.fw_rev);
		strncpy((char *)inq.venderId, tmp, sizeof(inq.venderId)+sizeof(inq.productId)+sizeof(inq.productVersion));
		if ( this->htype == HT_CDROM )
		{
			inq.devTypeModifier = 0x05;	// CD/DVD-ROM Device
			inq.removable       = 1;	// removable
			inq.ansiVer         = 1;
		//	inq.version |= 1;
		}
	}
	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	CD-ROM

int		DevCtl::cdrom_opendevice(void)
{
	int		res, stt;

	if ( (res = ioctl(this->fd, CDROM_DRIVE_STATUS, &stt)) < 0 )
		return -1;
	switch ( res )
	{
		case CDS_NO_INFO:
		case CDS_NO_DISC:
		case CDS_TRAY_OPEN:
		case CDS_DRIVE_NOT_READY:
		case CDS_DISC_OK:
			break;
	}
	this->blk.blkLen = 2048;
	this->blk.blkNum = 0;
	if ( res == CDS_DISC_OK )
	{
		this->blk.blkNum = lseek(this->fd, (off_t)0, SEEK_END) / this->blk.blkLen;
		lseek(this->fd, (off_t)0, SEEK_SET);
	}
	this->htype = HT_CDROM;
	return 0;
}

int		DevCtl::CdromReadToc(BOOL msf, unsigned size, void * buffer)
{
	int						res, i, trk;
	struct cdrom_tochdr		head;
	struct cdrom_tocentry	ent;
	SCSITOCDATA				*toc = (SCSITOCDATA *) buffer;

	if ( this->fd < 0 || this->htype != HT_CDROM || size < sizeof(*toc) || buffer == NULL )
		return -1;
	if ( (res = ioctl(this->fd, CDROMREADTOCHDR, &head)) < 0 )
		return -1;
	memset( buffer, 0, size);
	toc->startTrack = head.cdth_trk0;
	toc->lastTrack  = head.cdth_trk1;
	for ( trk = head.cdth_trk0, i = 0; trk <= head.cdth_trk1 && i < 100; trk++, i++ )
	{
		ent.cdte_track = trk;
		ent.cdte_format = msf ? CDROM_MSF : CDROM_LBA;
		if ( (res = ioctl(this->fd, CDROMREADTOCENTRY, &ent)) < 0 )
			break;
		toc->track[i].adrctl = (ent.cdte_adr<<4) | ent.cdte_ctrl;
		toc->track[i].no = trk;
		if ( msf )
		{
			toc->track[i].cdAddr[0] = 0;
			toc->track[i].cdAddr[1] = ent.cdte_addr.msf.minute;
			toc->track[i].cdAddr[2] = ent.cdte_addr.msf.second;
			toc->track[i].cdAddr[3] = ent.cdte_addr.msf.frame;
		} else
		{
			toc->track[i].cdAddr[0] = (BYTE)(ent.cdte_addr.lba>>24);
			toc->track[i].cdAddr[1] = (BYTE)(ent.cdte_addr.lba>>16);
			toc->track[i].cdAddr[2] = (BYTE)(ent.cdte_addr.lba>> 8);
			toc->track[i].cdAddr[3] = (BYTE)(ent.cdte_addr.lba    );
		}
	}
	if ( i < 100 )
	{
		ent.cdte_track = CDROM_LEADOUT;
		ent.cdte_format = msf ? CDROM_MSF : CDROM_LBA;
		if ( (res = ioctl(this->fd, CDROMREADTOCENTRY, &ent)) == 0 )
		{
			toc->track[i].adrctl = (ent.cdte_adr<<4) | ent.cdte_ctrl;
			toc->track[i].no = CDROM_LEADOUT;
			if ( msf )
			{
				toc->track[i].cdAddr[0] = 0;
				toc->track[i].cdAddr[1] = ent.cdte_addr.msf.minute;
				toc->track[i].cdAddr[2] = ent.cdte_addr.msf.second;
				toc->track[i].cdAddr[3] = ent.cdte_addr.msf.frame;
			} else
			{
				toc->track[i].cdAddr[0] = (BYTE)(ent.cdte_addr.lba>>24);
				toc->track[i].cdAddr[1] = (BYTE)(ent.cdte_addr.lba>>16);
				toc->track[i].cdAddr[2] = (BYTE)(ent.cdte_addr.lba>> 8);
				toc->track[i].cdAddr[3] = (BYTE)(ent.cdte_addr.lba    );
			}
		}
	}
		
	{
		unsigned	size;

		size = 10 + sizeof(*toc->track) * i;
		toc->tocLength[0] = (BYTE)(size>>8);
		toc->tocLength[1] = (BYTE)size;
	}
	
	return 0;
}
