//////////////////////////////////////////////////////////////////////////
//	"sdb_dbdev.cpp"
//------------------------------------------------------------------------
//	$Id: sdb_dbdev.cpp 104 2006-04-21 04:35:00Z taro $
//////////////////////////////////////////////////////////////////////////

#include	"../include/sdblib.h"
#include	"../lib/scrlib.h"
#if	defined(WIN32)
# include	<crtdbg.h>
#endif

#if	defined(_DEBUG) && defined(WIN32)
#	define	DBGREPORT(param_)	_CrtDbgReport param_
#else
#	define	DBGREPORT(param_)
# ifndef	_ASSERT
#	define	_ASSERT
# endif
#endif

namespace	sdb
{
SqlLib::TableItem	dev_tbl[DEV_FLD_NUM] =
{
	{ DEV_ID         , TEXT("dev_id"         ), SqlLib::DTYPE_SERIAL   , TEXT(""),   0, SqlLib::FF_SERIAL
	                                                                                  | SqlLib::FF_PRIMARY, TEXT("device-id"         )},
	{ DEV_MODEL      , TEXT("dev_model"      ), SqlLib::DTYPE_VARCHAR  , TEXT(""),  64, SqlLib::FF_INDEX  , TEXT("model"             )},
	{ DEV_SERIAL     , TEXT("dev_serial"     ), SqlLib::DTYPE_VARCHAR  , TEXT(""),  64, SqlLib::FF_INDEX  , TEXT("serial"            )},
	{ DEV_HOSTNAME   , TEXT("dev_hostname"   ), SqlLib::DTYPE_VARCHAR  , TEXT(""),  64, SqlLib::FF_INDEX  , TEXT("hostname"          )},
	{ DEV_GROUPNAME  , TEXT("dev_groupname"  ), SqlLib::DTYPE_VARCHAR  , TEXT(""),  64, SqlLib::FF_INDEX  , TEXT("groupname"         )},
	{ DEV_SMT_TBLNAME, TEXT("dev_smt_tblname"), SqlLib::DTYPE_TEXT     , TEXT(""),  64, 0                 , TEXT("table name"        )},
	{ DEV_F_DISABLE  , TEXT("dev_f_disable"  ), SqlLib::DTYPE_BOOLEAN  , TEXT(""),   0, 0                 , TEXT("disable"           )},
	{ DEV_BLOCKSIZE  , TEXT("dev_blocksize"  ), SqlLib::DTYPE_INTEGER  , TEXT(""),   0, 0                 , TEXT("block-size"        )},
	{ DEV_BLOCKNUMBER, TEXT("dev_blocknumber"), SqlLib::DTYPE_BIGINT   , TEXT(""),   0, 0                 , TEXT("block-number"      )},
	{ DEV_TS_BEGIN   , TEXT("dev_ts_begin"   ), SqlLib::DTYPE_TIMESTAMP, TEXT(""),   0, 0                 , TEXT("begin date-time"   )},
	{ DEV_TS_LAST    , TEXT("dev_ts_last"    ), SqlLib::DTYPE_TIMESTAMP, TEXT(""),   0, 0                 , TEXT("last date-time"    )},
	{ DEV_USER_STT   , TEXT("dev_user_stt"   ), SqlLib::DTYPE_INTEGER  , TEXT(""),   0, 0                 , TEXT("user status"       )},
	{ DEV_USER_ID    , TEXT("dev_user_id"    ), SqlLib::DTYPE_INTEGER  , TEXT(""),   0, 0                 , TEXT("user id"           )},
	{ DEV_USER_PASS  , TEXT("dev_user_pass"  ), SqlLib::DTYPE_VARCHAR  , TEXT(""),  64, 0                 , TEXT("user pass"         )},
	{ DEV_USER_TS    , TEXT("dev_user_ts"    ), SqlLib::DTYPE_TIMESTAMP, TEXT(""),   0, 0                 , TEXT("user date-time"    )},
	{ DEV_MEMO       , TEXT("dev_memo"       ), SqlLib::DTYPE_TEXT     , TEXT(""),   0, 0                 , TEXT("remark"            )},
};

DBDevLib::DBDevLib()
{
	this->tblname.SetText(SDB_DEV_TBLNAME);

	{
		int				i;

		for ( i = 0; i < sizeof(dev_tbl)/sizeof(*dev_tbl); i++ )
		{
			if ( this->tbldef.Add(dev_tbl[i]) < 0 )
			{
				_ASSERT(0);
				break;	// error
			}
		}
	}
	return;
}

DBDevLib::~DBDevLib()
{
	return;
}

int			DBDevLib::CreateTable(void)
{
	SqlLib::VarBuffer	addtbl;

	if ( this->tblname.GetLength() == 0 || this->tbldef.GetCount() == 0 )
		return -1;
	addtbl.AddText(TEXT("UNIQUE ("));
	addtbl.AddText(dev_tbl[DEV_MODEL].fname);
	addtbl.AddText(TEXT(","));
	addtbl.AddText(dev_tbl[DEV_SERIAL].fname);
	addtbl.AddText(TEXT(")"));
	addtbl.AddNulChar();
	return this->DBTblLib::CreateTable(this->tblname.GetText(), addtbl.GetText());
}

int			DBDevLib::SearchDevice(SqlLib::FieldsTbl & data, const TCHAR * model, const TCHAR * serial)
{
	SqlLib				*sqllib;
	SqlLib::VarBuffer	sql;
	SqlLib::Result		res;
	int					ymax;

	_ASSERT(this->sdblib != NULL);
	sqllib = this->sdblib->sqllib;
	_ASSERT(sqllib != NULL);

	sql.AddText(TEXT("SELECT * FROM "));
	sql.AddText(this->tblname.GetText());
	sql.AddText(TEXT(" WHERE "));
	sql.AddText( sdb::dev_tbl[sdb::DEV_MODEL].fname );
	sql.AddText(TEXT("="));
	SqlLib::EscapeString(sql, model , TEXT('\''));
	sql.AddText(TEXT(" AND "));
	sql.AddText( sdb::dev_tbl[sdb::DEV_SERIAL].fname );
	sql.AddText(TEXT("="));
	SqlLib::EscapeString(sql, serial, TEXT('\''));
	sql.AddNulChar();

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"sql = %s\n", sql.GetText()));
	sqllib->Begin();
	if ( sqllib->Exec(res, sql.GetText()) < 0 || sqllib->GetStatus(res) != SqlLib::RES_OK )
	{
		sqllib->Rollback();
		return -1;
	}
	sqllib->Commit();

	data.Clear();
	if ( (ymax = sqllib->GetTuples(res)) > 0  )
	{
		int		y, x, xmax;

		y = 0;
		if ( (xmax = sqllib->GetFields(res)) > 0 )
		{
			for ( x = 0; x < xmax; x++ )
			{
				const TCHAR			*fname;
				SqlLib::DataValue	val;

				if ( (fname  = sqllib->GetFname(res, x)) != NULL )
				{
					if ( sqllib->GetDataValue(res, y, x, val) >= 0 )
						data.Add(fname, val);
				}
			}
		}
	}
	sqllib->ClearResult(res);

	return ymax;
}

int			DBDevLib::SearchDevice(DBDevData & data, const TCHAR * model, const TCHAR * serial)
{
	SqlLib::FieldsTbl	fdata;
	int					ymax;

	if ( (ymax = this->SearchDevice(fdata, model, serial)) < 1 )
		return -1;
	DBDevLib::GetDevData(data, fdata);
	return ymax;
}

int			DBDevLib::GetDevData(DBDevData & data, SqlLib::FieldsTbl & fdata)
{
	int			x;

	for ( x = 0; x < fdata.GetCount(); x++ )
	{
		SqlLib::DataValue	val;
		const TCHAR			*fname;
		int					fno;

		if ( (fname = fdata.GetValue(x, val)) == NULL )
			continue;
		fno = this->tbldef.GetFno(fname);
		if ( fno < 0 || fno >= DEV_FLD_NUM )
		{
			continue;
		} else if ( dev_tbl[fno].dtype != val.GetDataType() )
		{
			if ( (dev_tbl[fno].flags & SqlLib::FF_SERIAL) && val.GetDataType() == SqlLib::DTYPE_INTEGER )
			{
				;
			} else
				continue;
		}
		switch ( fno )
		{
		case DEV_ID         : data.dev_id                = val.GetInteger()  ; break;
		case DEV_MODEL      : data.dev_model.SetText(      val.GetVarChar() ); break;
		case DEV_SERIAL     : data.dev_serial.SetText(     val.GetVarChar() ); break;
		case DEV_HOSTNAME   : data.dev_hostname.SetText(   val.GetVarChar() ); break;
		case DEV_GROUPNAME  : data.dev_groupname.SetText(  val.GetVarChar() ); break;
		case DEV_SMT_TBLNAME: data.dev_smt_tblname.SetText(val.GetVarChar() ); break;
		case DEV_F_DISABLE  : data.dev_f_disable         = val.GetBoolean()  ; break;
		case DEV_BLOCKSIZE  : data.dev_blockSize         = val.GetInteger()  ; break;
		case DEV_BLOCKNUMBER: data.dev_blockNumber       = val.GetBigInt()   ; break;
		case DEV_TS_BEGIN   : data.dev_ts_begin          = val.GetTimeStamp(); break;
		case DEV_TS_LAST    : data.dev_ts_last           = val.GetTimeStamp(); break;
		case DEV_USER_STT   : data.dev_user_stt          = val.GetInteger()  ; break;
		case DEV_USER_ID    : data.dev_user_id           = val.GetInteger()  ; break;
		case DEV_USER_PASS  : data.dev_user_pass.SetText(  val.GetText()    ); break;
		case DEV_USER_TS    : data.dev_user_ts           = val.GetTimeStamp(); break;
		case DEV_MEMO       : data.dev_memo.SetText(       val.GetText()    ); break;
		}
	}
	return 0;
}

int			DBDevLib::SetDevData(SqlLib::FieldsTbl & fdata, const DBDevData & data)
{
	fdata.Clear();
//	skip DEV_ID
	fdata.AddVarChar(  dev_tbl[DEV_MODEL      ].fname, data.dev_model.GetText()      , dev_tbl[DEV_MODEL      ].size);
	fdata.AddVarChar(  dev_tbl[DEV_SERIAL     ].fname, data.dev_serial.GetText()     , dev_tbl[DEV_SERIAL     ].size);
	fdata.AddVarChar(  dev_tbl[DEV_HOSTNAME   ].fname, data.dev_hostname.GetText()   , dev_tbl[DEV_HOSTNAME   ].size);
	fdata.AddVarChar(  dev_tbl[DEV_GROUPNAME  ].fname, data.dev_groupname.GetText()  , dev_tbl[DEV_GROUPNAME  ].size);
	fdata.AddVarChar(  dev_tbl[DEV_SMT_TBLNAME].fname, data.dev_smt_tblname.GetText(), dev_tbl[DEV_SMT_TBLNAME].size);
	fdata.AddBoolean(  dev_tbl[DEV_F_DISABLE  ].fname, data.dev_f_disable);
	fdata.AddInteger(  dev_tbl[DEV_BLOCKSIZE  ].fname, data.dev_blockSize);
	fdata.AddBigInt(   dev_tbl[DEV_BLOCKNUMBER].fname, data.dev_blockNumber);
	fdata.AddTimeStamp(dev_tbl[DEV_TS_BEGIN   ].fname, data.dev_ts_begin);
	fdata.AddTimeStamp(dev_tbl[DEV_TS_LAST    ].fname, data.dev_ts_last);
	fdata.AddInteger(  dev_tbl[DEV_USER_STT   ].fname, data.dev_user_stt);
	fdata.AddInteger(  dev_tbl[DEV_USER_ID    ].fname, data.dev_user_id);
	fdata.AddText(     dev_tbl[DEV_USER_PASS  ].fname, data.dev_user_pass.GetText());
	fdata.AddTimeStamp(dev_tbl[DEV_USER_TS    ].fname, data.dev_user_ts);
	fdata.AddText(     dev_tbl[DEV_MEMO       ].fname, data.dev_memo.GetText());

	return 0;
}

BOOL		DBDevLib::GetDataValue(SqlLib::DataValue & val, SqlLib::FieldsTbl & data, DevFldNo no) // static
{
	int					i;
	const TCHAR			*fname;

	if ( (int)no < 0 || no > DEV_FLD_NUM )
		return FALSE;
	for ( i = 0; i < data.GetCount(); i++ )
	{
		if ( (fname = data.GetValue(i, val)) != NULL )
		{
			if ( SQStr::StrCmpi(fname, sdb::dev_tbl[no].fname) == 0 )
				return TRUE;
		}
	}
	return FALSE;
}

int		DBDevLib::AddDevice(SqlLib::FieldsTbl & fdata, const DBDevData & data)
{
	TCHAR				query[1024];
	SqlLib::Result		res;
	SqlLib::Oid			oid;
	SqlLib::DataValue	val;
	SqlLib::BigInt		lastval;

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"DBDevLib::AddDevice()\n"));
	if ( this->sdblib->dbdev.SetDevData(fdata, data) < 0 )
		return -1;
	this->sdblib->sqllib->Begin();
	if ( (oid = this->sdblib->dbdev.InsertData(fdata)) > 0 )
	{
		_stprintf(query, TEXT("SELECT %s FROM %s WHERE OID=%d")
		  , dev_tbl[DEV_ID].fname
		  , this->tblname.GetText()
		  , oid);
	} if ( oid == 0 && this->sdblib->sqllib->GetLastVal(lastval) > 0 )
	{
		_stprintf(query, TEXT("SELECT %s FROM %s WHERE %s=%"FI64"d")
		  , dev_tbl[DEV_ID].fname
		  , this->tblname.GetText()
		  , dev_tbl[DEV_ID].fname
		  , lastval);
	}  else
	{
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"sql error (oid = %d)\n", (int)oid));
		this->sdblib->sqllib->Rollback();
		return -1;
	}
	this->sdblib->sqllib->Commit();
	DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"query = %s\n", query));
	this->sdblib->sqllib->Begin();
	if ( this->sdblib->sqllib->Exec(res, query) < 0 || this->sdblib->sqllib->GetStatus(res) != SqlLib::RES_OK )
	{
		this->sdblib->sqllib->Rollback();
		return -1;
	}
	this->sdblib->sqllib->Commit();

	if ( this->sdblib->sqllib->GetDataValue(res, 0, 0, val) < 0 )
		return -1;
	DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"dev_id = %d\n", val.GetInteger()));
	fdata.AddInteger(dev_tbl[DEV_ID].fname, val.GetInteger());

	return 0;
}

// Need: data.dev_id
int		DBDevLib::DeleteDevice(const DBDevData & devdata)
{
	TCHAR				where[80];
	SqlLib::Result		res;

#if	defined(_DEBUG) && defined(SDBCTL)
	_ftprintf(stderr, TEXT("DBDevLib::DeleteDevice(dev_id = %d)\n"), devdata.dev_id);
#endif
	this->sdblib->sqllib->Begin();
	_stprintf(where, TEXT("%s=%d"), sdb::dev_tbl[sdb::DEV_ID].fname, devdata.dev_id);
#if	defined(_DEBUG) && defined(SDBCTL)
	_ftprintf(stderr, TEXT("  where  %s\n"), where);
#endif
	if ( this->sdblib->dbdev.DeleteData(where) < 0 )
	{
		this->sdblib->sqllib->Rollback();
		return -1;
	}
	this->sdblib->sqllib->Commit();

	return 0;
}

int			DBDevLib::GetDevDataList(DBDevDataList & list, const TCHAR * add_sql)
{
	SqlLib				*sqllib = this->GetSqlLib();
	SqlLib::VarBuffer	sql;
	SqlLib::Result		res;

	sql.AddText(TEXT("SELECT * FROM "));
	sql.AddText(this->tblname.GetText());
	if ( add_sql != NULL )
	{
		sql.AddText(TEXT(" "));
		sql.AddText(add_sql);
	}
	sql.AddNulChar();

	sqllib->Begin();
	if ( sqllib->Exec(res, sql.GetText()) < 0 || sqllib->GetStatus(res) != SqlLib::RES_OK )
	{
		sqllib->Rollback();
		return -1;
	}
	sqllib->Commit();

	list.clear();
	{
		int					x, y, xmax, ymax;
		SqlLib::FieldsTbl	data;

		ymax = sqllib->GetTuples(res);
		xmax = sqllib->GetFields(res);
		for ( y = 0; y < ymax; y++ )
		{
			data.Clear();
			for ( x = 0; x < xmax; x++ )
			{
				const TCHAR			*fname;
				SqlLib::DataValue	val;

				if ( (fname  = sqllib->GetFname(res, x)) != NULL )
				{
					if ( sqllib->GetDataValue(res, y, x, val) >= 0 )
						data.Add(fname, val);
				}
			}
			{
				DBDevData			devdata;

				this->GetDevData(devdata, data);
				list.push_back(devdata);
			}
		}
	}
	sqllib->ClearResult(res);

	return 0;
}

};// end of namespace sdb;
