//////////////////////////////////////////////////////////////////////////
//	"sdb_dbsmt.cpp"
//------------------------------------------------------------------------
//	$Id: sdb_dbsmt.cpp 70 2006-03-28 02:58:31Z taro $
//////////////////////////////////////////////////////////////////////////

#include	"../include/sdblib.h"
#include	"../lib/scrlib.h"
#if	defined(WIN32)
# include	<crtdbg.h>
#endif

#if	defined(_DEBUG) && defined(WIN32)
#	define	DBGREPORT(param_)	_CrtDbgReport param_
#else
#	define	DBGREPORT(param_)
# ifndef	_ASSERT
#	define	_ASSERT
# endif
#endif

namespace	sdb
{
SqlLib::TableItem	smt_tbl[SMT_FLD_NUM] =
{
	{ SMT_TS       , TEXT("smt_ts"         ), SqlLib::DTYPE_TIMESTAMP, TEXT(""),   0, SqlLib::FF_INDEX
	                                                                                | SqlLib::FF_PRIMARY, TEXT("date-time"         )},

	{ SMT_DEV_ID   , TEXT("smt_dev_id"     ), SqlLib::DTYPE_INTEGER  , TEXT(""),   0, SqlLib::FF_INDEX
	                                                                                | SqlLib::FF_PRIMARY, TEXT("device-id"        )},

	{ SMT_ATTR_ID  , TEXT("smt_attr_id"    ), SqlLib::DTYPE_SMALLINT , TEXT(""),   0, SqlLib::FF_PRIMARY, TEXT("attribute id#"     )},
	{ SMT_VALUE    , TEXT("smt_value"      ), SqlLib::DTYPE_SMALLINT , TEXT(""),   0, 0                 , TEXT("value"             )},
	{ SMT_WORST    , TEXT("smt_worst"      ), SqlLib::DTYPE_SMALLINT , TEXT(""),   0, 0                 , TEXT("worst"             )},
	{ SMT_THRESHOLD, TEXT("smt_threshold"  ), SqlLib::DTYPE_SMALLINT , TEXT(""),   0, 0                 , TEXT("threshold"         )},
	{ SMT_FLAGS    , TEXT("smt_flags"      ), SqlLib::DTYPE_SMALLINT , TEXT(""),   0, 0                 , TEXT("flags"             )},
	{ SMT_RAW_VALUE, TEXT("smt_raw_value"  ), SqlLib::DTYPE_BIGINT   , TEXT(""),   0, 0                 , TEXT("raw value"         )},
	{ SMT_RAW_DATA , TEXT("smt_raw_data"   ), SqlLib::DTYPE_BIGINT   , TEXT(""),   0, 0                 , TEXT("raw data"          )},
};

DBSmtLib::DBSmtLib()
{
	this->tblname.SetText(SDB_DEFAULT_SMT_TBLNAME);

	{
		int				i;

		for ( i = 0; i < sizeof(smt_tbl)/sizeof(*smt_tbl); i++ )
		{
			if ( this->tbldef.Add(smt_tbl[i]) < 0 )
			{
				_ASSERT(0);
				break;	// error
			}
		}
	}
	return;
}

DBSmtLib::~DBSmtLib()
{
	return;
}

int			DBSmtLib::CreateTable(void)
{
	if ( this->tblname.GetLength() == 0 || this->tbldef.GetCount() == 0 )
		return -1;
	return this->DBTblLib::CreateTable(this->tblname.GetText());
}

int			DBSmtLib::GetSmtData(DBSmtData & data, SqlLib::Result & sqlres, int y, int xmax, const TCHAR ** fnametbl)
{
	int			x;

	_ASSERT(this->sdblib != NULL && this->sdblib->sqllib != NULL);
	if ( xmax == 0 )
		xmax = this->sdblib->sqllib->GetFields(sqlres);
	for ( x = 0; x < xmax; x++ )
	{
		const TCHAR			*fname;
		int					fno;
		SqlLib::DataValue	val;

		if ( this->sdblib->sqllib->GetDataValue(sqlres, y, x, val) < 0 )
			continue;	// error
		if ( fnametbl == NULL )
			fname = this->sdblib->sqllib->GetFname(sqlres, x);
		else
			fname = fnametbl[x];
		fno = this->tbldef.GetFno(fname);
		if ( fno < 0 || fno >= SMT_FLD_NUM )
		{
			continue;
		} else if ( smt_tbl[fno].dtype != val.GetDataType() )
		{
			if ( (smt_tbl[fno].flags & SqlLib::FF_SERIAL) && val.GetDataType() == SqlLib::DTYPE_INTEGER )
			{
				;
			} else
				continue;
		}
		switch ( fno )
		{
		case SMT_TS         : data.smt_ts          = val.GetTimeStamp(); break;
		case SMT_DEV_ID     : data.smt_dev_id      = val.GetInteger()  ; break;
		case SMT_ATTR_ID    : data.smt_attr_id     = val.GetSmallInt() ; break;
		case SMT_VALUE      : data.smt_value       = val.GetSmallInt() ; break;
		case SMT_WORST      : data.smt_worst       = val.GetSmallInt() ; break;
		case SMT_THRESHOLD  : data.smt_threshold   = val.GetSmallInt() ; break;
		case SMT_FLAGS      : data.smt_flags       = val.GetSmallInt() ; break;
		case SMT_RAW_VALUE  : data.smt_raw_value   = val.GetBigInt()   ; break;
		case SMT_RAW_DATA   :
		{
			sdb::Int64	raw;

			raw = val.GetBigInt();
			sdb::ATA_SET48(data.smt_raw_data, raw);
			break;
		}
		} // end of switch
	} // end of for(x;;)
	return 0;
}

int			DBSmtLib::GetSmtData(DBSmtData & data, SqlLib::FieldsTbl & fdata)
{
	int			x;

	for ( x = 0; x < fdata.GetCount(); x++ )
	{
		SqlLib::DataValue	val;
		const TCHAR			*fname;
		int					fno;

		if ( (fname = fdata.GetValue(x, val)) == NULL )
			continue;
		fno = this->tbldef.GetFno(fname);
		if ( fno < 0 || fno >= SMT_FLD_NUM )
		{
			continue;
		} else if ( smt_tbl[fno].dtype != val.GetDataType() )
		{
			if ( (smt_tbl[fno].flags & SqlLib::FF_SERIAL) && val.GetDataType() == SqlLib::DTYPE_INTEGER )
			{
				;
			} else
				continue;
		}
		switch ( fno )
		{
		case SMT_TS         : data.smt_ts                = val.GetTimeStamp(); break;
		case SMT_DEV_ID     : data.smt_dev_id            = val.GetInteger()  ; break;
		case SMT_ATTR_ID    : data.smt_attr_id           = val.GetSmallInt() ; break;
		case SMT_VALUE      : data.smt_value             = val.GetSmallInt() ; break;
		case SMT_WORST      : data.smt_worst             = val.GetSmallInt() ; break;
		case SMT_THRESHOLD  : data.smt_threshold         = val.GetSmallInt() ; break;
		case SMT_FLAGS      : data.smt_flags             = val.GetSmallInt() ; break;
		case SMT_RAW_VALUE  : data.smt_raw_value         = val.GetBigInt()   ; break;
		case SMT_RAW_DATA   :
		{
			sdb::Int64	raw;

			raw = val.GetBigInt();
			sdb::ATA_SET48(data.smt_raw_data, raw);
			break;
		}
		}
	}
	return 0;
}

int			DBSmtLib::SetSmtData(SqlLib::FieldsTbl & fdata, const DBSmtData & data)
{
	fdata.Clear();
	fdata.AddTimeStamp(smt_tbl[SMT_TS       ].fname, data.smt_ts       );
	fdata.AddInteger(  smt_tbl[SMT_DEV_ID   ].fname, data.smt_dev_id   );
	fdata.AddSmallInt( smt_tbl[SMT_ATTR_ID  ].fname, data.smt_attr_id  );
	fdata.AddSmallInt( smt_tbl[SMT_VALUE    ].fname, data.smt_value    );
	fdata.AddSmallInt( smt_tbl[SMT_WORST    ].fname, data.smt_value    );
	fdata.AddSmallInt( smt_tbl[SMT_THRESHOLD].fname, data.smt_threshold);
	fdata.AddSmallInt( smt_tbl[SMT_FLAGS    ].fname, data.smt_flags    );
	fdata.AddBigInt  ( smt_tbl[SMT_RAW_VALUE].fname, data.smt_raw_value);
	fdata.AddBigInt  ( smt_tbl[SMT_RAW_DATA].fname , sdb::ATA_GET48((void *)data.smt_raw_data));

	return 0;
}


};// end of namespace sdb;
