//////////////////////////////////////////////////////////////////////////
//	"sdb_dbwat.cpp"
//------------------------------------------------------------------------
//	$Id: sdb_dbwat.cpp 70 2006-03-28 02:58:31Z taro $
//////////////////////////////////////////////////////////////////////////

#include	"../include/sdblib.h"
#include	"../lib/scrlib.h"
#if	defined(WIN32)
# include	<crtdbg.h>
#endif

#if	defined(_DEBUG) && defined(WIN32)
#	define	DBGREPORT(param_)	_CrtDbgReport param_
#else
#	define	DBGREPORT(param_)
# ifndef	_ASSERT
#	define	_ASSERT
# endif
#endif

namespace	sdb
{
SqlLib::TableItem	wat_tbl[WAT_FLD_NUM] =
{
	{ WAT_ID       , TEXT("wat_id"         ), SqlLib::DTYPE_SERIAL   , TEXT(""),   0, SqlLib::FF_SERIAL
	                                                                                | SqlLib::FF_PRIMARY, TEXT("watch-id"          )},
	{ WAT_PRIORITY , TEXT("wat_priority"   ), SqlLib::DTYPE_SMALLINT , TEXT(""),   0, 0                 , TEXT("priority"          )},
	{ WAT_MODEL    , TEXT("wat_model"      ), SqlLib::DTYPE_VARCHAR  , TEXT(""),  64, 0                 , TEXT("model"             )},
	{ WAT_SERIAL   , TEXT("wat_serial"     ), SqlLib::DTYPE_VARCHAR  , TEXT(""),  64, 0                 , TEXT("serial"            )},
	{ WAT_ATTR_ID  , TEXT("wat_attr_id"    ), SqlLib::DTYPE_SMALLINT , TEXT(""),   0, 0                 , TEXT("attribute id"      )},
	{ WAT_TYPE     , TEXT("wat_type"       ), SqlLib::DTYPE_SMALLINT , TEXT(""),   0, 0                 , TEXT("watch type"        )},
	{ WAT_F_DISABLE, TEXT("wat_f_disable"  ), SqlLib::DTYPE_BOOLEAN  , TEXT(""),   0, 0                 , TEXT("disable flag"      )},
	{ WAT_WARN     , TEXT("wat_warn"       ), SqlLib::DTYPE_SMALLINT , TEXT(""),   0, 0                 , TEXT("warnning level"    )},
	{ WAT_OP       , TEXT("wat_op"         ), SqlLib::DTYPE_VARCHAR  , TEXT(""),  32, 0                 , TEXT("operator"          )},
	{ WAT_RIGHT    , TEXT("wat_right"      ), SqlLib::DTYPE_VARCHAR  , TEXT(""),  64, 0                 , TEXT("right value"       )},
};

DBWatLib::DBWatLib()
{
	this->tblname.SetText(SDB_WAT_TBLNAME);

	{
		int				i;

		for ( i = 0; i < sizeof(wat_tbl)/sizeof(*wat_tbl); i++ )
		{
			if ( this->tbldef.Add(wat_tbl[i]) < 0 )
			{
				_ASSERT(0);
				break;	// error
			}
		}
	}
	return;
}

DBWatLib::~DBWatLib()
{
	return;
}

DBWatLib *	DBWatLib::Create(SdbLib * sdblib) // static
{
	DBWatLib *	dbwat;

	if ( (dbwat = new DBWatLib) != NULL )
		dbwat->sdblib = sdblib;

	return dbwat;
}

int			DBWatLib::CreateTable(void)
{
	if ( this->tblname.GetLength() == 0 || this->tbldef.GetCount() == 0 )
		return -1;
	return this->DBTblLib::CreateTable(this->tblname.GetText());
}

int			DBWatLib::GetWatData(DBWatData & data, SqlLib::FieldsTbl & fdata)
{
	int			x;

	for ( x = 0; x < fdata.GetCount(); x++ )
	{
		SqlLib::DataValue	val;
		const TCHAR			*fname;
		int					fno;

		if ( (fname = fdata.GetValue(x, val)) == NULL )
			continue;
		fno = this->tbldef.GetFno(fname);
		if ( fno < 0 || fno >= WAT_FLD_NUM )
		{
			continue;
		} else if ( wat_tbl[fno].dtype != val.GetDataType() )
		{
			if ( (wat_tbl[fno].flags & SqlLib::FF_SERIAL) && val.GetDataType() == SqlLib::DTYPE_INTEGER )
			{
				;
			} else
				continue;
		}
		switch ( fno )
		{
		case WAT_ID         : data.wat_id                = val.GetInteger()  ; break;
		case WAT_PRIORITY   : data.wat_priority          = val.GetSmallInt() ; break;
		case WAT_MODEL      : data.wat_model.SetText(      val.GetVarChar()) ; break;
		case WAT_SERIAL     : data.wat_serial.SetText(     val.GetVarChar()) ; break;
		case WAT_ATTR_ID    : data.wat_attr_id           = val.GetSmallInt() ; break;
		case WAT_TYPE       : data.wat_type = (sdb::WatchType)val.GetSmallInt() ; break;
		case WAT_F_DISABLE  : data.wat_f_disable         = val.GetBoolean()  ; break;
		case WAT_WARN       : data.wat_warn              = val.GetSmallInt() ; break;
		case WAT_OP         : data.wat_op.SetText(        val.GetVarChar()) ; break;
		case WAT_RIGHT      : data.wat_right.SetText(     val.GetVarChar()) ; break;
		}
	}
	return 0;
}

int			DBWatLib::SetWatData(SqlLib::FieldsTbl & fdata, const DBWatData & data)
{
	fdata.Clear();
//	skip wat_id
	fdata.AddSmallInt( wat_tbl[WAT_PRIORITY ].fname, data.wat_priority );
	fdata.AddVarChar(  wat_tbl[WAT_MODEL    ].fname, data.wat_model.GetText()   , wat_tbl[WAT_MODEL      ].size);
	fdata.AddVarChar(  wat_tbl[WAT_SERIAL   ].fname, data.wat_serial.GetText()  , wat_tbl[WAT_SERIAL     ].size);
	fdata.AddSmallInt( wat_tbl[WAT_ATTR_ID  ].fname, data.wat_attr_id   );
	fdata.AddSmallInt( wat_tbl[WAT_TYPE     ].fname, data.wat_type      );
	fdata.AddBoolean(  wat_tbl[WAT_F_DISABLE].fname, data.wat_f_disable );
	fdata.AddSmallInt( wat_tbl[WAT_WARN     ].fname, data.wat_warn      );
	fdata.AddVarChar(  wat_tbl[WAT_OP       ].fname, data.wat_op.GetText()      , wat_tbl[WAT_OP         ].size);
	fdata.AddVarChar(  wat_tbl[WAT_RIGHT    ].fname, data.wat_right.GetText()   , wat_tbl[WAT_RIGHT      ].size);

	return 0;
}


};// end of namespace sdb;
