//////////////////////////////////////////////////////////////////////////
//	"sdblib.cpp"
//------------------------------------------------------------------------
//	$Id: sdblib.cpp 104 2006-04-21 04:35:00Z taro $
//////////////////////////////////////////////////////////////////////////

#include	"../include/sdblib.h"
#include	"../lib/scrlib.h"
#include	"../lib/varbuf.h"
#include	"../lib/tform.h"
#if	defined(WIN32)
# include	<crtdbg.h>
#else
# include	<time.h>
#endif

#if	defined(_DEBUG) && defined(WIN32)
#	define	DBGREPORT(param_)	_CrtDbgReport param_
#else
#	define	DBGREPORT(param_)
# ifndef	_ASSERT
#	define	_ASSERT
# endif
#endif

namespace	sdb
{
const TCHAR				*db_host = SDB_DEFAULT_HOST;
const TCHAR				*db_port = SDB_DEFAULT_PORT;
const TCHAR				*db_user = SDB_DEFAULT_USER;
const TCHAR				*db_pass = SDB_DEFAULT_PASS;
const TCHAR				*db_name = SDB_DEFAULT_DBNAME;

const TCHAR				*sub_dir = NULL;

const TCHAR				*http_post_host = NULL;
const TCHAR				*http_post_port = NULL;
const TCHAR				*http_post_path = NULL;
const TCHAR				*http_auth_type = NULL;
const TCHAR				*http_auth_user = NULL;
const TCHAR				*http_auth_pass = NULL;

int						mail_warnlevel    = 0;
const TCHAR				*mail_from        = NULL;
const TCHAR				*mail_to          = NULL;
const TCHAR				*mail_smtp_server = NULL;

int						warn_high_estimate =  3;
int						warn_mid_estimate  = 14;
int						warn_low_estimate  = 30;
int						watch_offline_margin_sec = 2 * 60 * 60;

int						auto_raw_word_id      = 9;	// Power-ON Hour
int						random_wait_sec       = SDB_RANDOM_WAIT_SEC;
int						dialog_auto_close_sec = SDB_DIALOG_AUTO_CLOSE_SEC;

//////////////////////////////////////////////////////////////////////////
//	DBDevData

DBDevData::DBDevData()
{
}

DBDevData::DBDevData(const DBDevData & src)
{
	*this = src;
}

DBDevData & 		DBDevData::operator = (const DBDevData & src)
{
	this->dev_id          = src.dev_id;
	this->dev_model       = src.dev_model;
	this->dev_serial      = src.dev_serial;
	this->dev_hostname    = src.dev_hostname;
	this->dev_groupname   = src.dev_groupname;
	this->dev_smt_tblname = src.dev_smt_tblname;
	this->dev_f_disable   = src.dev_f_disable;
	this->dev_blockSize   = src.dev_blockSize;
	this->dev_blockNumber = src.dev_blockNumber;
	this->dev_ts_begin    = src.dev_ts_begin;
	this->dev_ts_last     = src.dev_ts_last;
	this->dev_user_stt    = src.dev_user_stt;
	this->dev_user_id     = src.dev_user_id;
	this->dev_user_pass   = src.dev_user_pass;
	this->dev_user_ts     = src.dev_user_ts;
	this->dev_memo        = src.dev_memo;

	return *this;
}

//////////////////////////////////////////////////////////////////////////
//	DBSmtData

DBSmtData::DBSmtData()
{
}

DBSmtData::DBSmtData(const DBSmtData & src)
{
	*this = src;
}

DBSmtData & 		DBSmtData::operator = (const DBSmtData & src)
{
	this->smt_ts          = src.smt_ts;
	this->smt_dev_id      = src.smt_dev_id;
	this->smt_attr_id     = src.smt_attr_id;
	this->smt_value       = src.smt_value;
	this->smt_worst       = src.smt_worst;
	this->smt_threshold   = src.smt_threshold;
	this->smt_flags       = src.smt_flags;
	this->smt_raw_value   = src.smt_raw_value;
	memcpy(this->smt_raw_data, src.smt_raw_data, sizeof(this->smt_raw_data));

	return *this;
}

//////////////////////////////////////////////////////////////////////////
//	DBWatData

DBWatData::DBWatData()
{
}

DBWatData::DBWatData(const DBWatData & src)
{
	*this = src;
}

DBWatData & 		DBWatData::operator = (const DBWatData & src)
{
	this->wat_id          = src.wat_id;
	this->wat_priority    = src.wat_priority;
	this->wat_model       = src.wat_model;
	this->wat_serial      = src.wat_serial;
	this->wat_attr_id     = src.wat_attr_id;
	this->wat_type        = src.wat_type;
	this->wat_f_disable   = src.wat_f_disable;
	this->wat_warn        = src.wat_warn;
	this->wat_op          = src.wat_op;
	this->wat_right       = src.wat_right;

	return *this;
}

//////////////////////////////////////////////////////////////////////////
//	DBTblLib

DBTblLib::DBTblLib()
{
	this->sdblib = NULL;

	return;
}

DBTblLib::~DBTblLib()
{
	return;
}

SqlLib *	DBTblLib::GetSqlLib(void) const
{
	return this->sdblib != NULL ? this->sdblib->sqllib : NULL;
}

int		DBTblLib::CreateTable(const TCHAR * tblname, const TCHAR * addtbl)
{
	_ASSERT(this->sdblib != NULL && this->sdblib->sqllib != NULL);
	if ( tblname == NULL )
	{
		if ( this->tblname.GetLength() == 0 )
			return -1;
		tblname = this->tblname.GetText();
	}
	if ( tblname == NULL || *tblname == TEXT('\0') )
		return -1;
	if ( this->sdblib->sqllib->CreateTable(tblname, this->tbldef, addtbl) < 0 )
		return -1;

	return 0;
}

int		DBTblLib::DropTable(const TCHAR * tblname)
{
	_ASSERT(this->sdblib != NULL && this->sdblib->sqllib != NULL);
	if ( tblname == NULL )
	{
		if ( this->tblname.GetLength() == 0 )
			return -1;
		tblname = this->tblname.GetText();
	}
	if ( tblname == NULL || *tblname == TEXT('\0') )
		return -1;
	if ( this->sdblib->sqllib->DropTable(tblname, this->tbldef) < 0 )
		return -1;

	return 0;
}

SqlLib::Oid	DBTblLib::InsertData(SqlLib::FieldsTbl & data)
{
	SqlLib::Oid			oid = -1;
	SqlLib::VarBuffer	sql;
	SqlLib::Result		sqlres;

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"DBTblLib::InsertData()\n"));
	_ASSERT(this->sdblib != NULL && this->sdblib->sqllib != NULL);
	_ASSERT(this->tblname.GetLength() > 0);
	if ( this->sdblib->sqllib->GetSql_Insert(sql, this->tblname.GetText(), this->tbldef, data) < 0 )
		return -1;

	this->sdblib->sqllib->Begin();
	if ( this->sdblib->sqllib->Exec(sqlres, sql.GetText()) < 0
	  || this->sdblib->sqllib->GetStatus(sqlres) != SqlLib::RES_OK )
	{
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"DBTblLib::InsertData: Exec or GetStatus Error\n"));
		this->sdblib->sqllib->Rollback();
		return -1;
	}
	oid = this->sdblib->sqllib->GetOidValue(sqlres);
	this->sdblib->sqllib->Commit();

	return oid;
}

int			DBTblLib::UpdateData(SqlLib::FieldsTbl & data, const TCHAR * where)
{
	SqlLib::VarBuffer	sql;
	SqlLib::Result		sqlres;

	_ASSERT(this->sdblib != NULL && this->sdblib->sqllib != NULL);
	_ASSERT(this->tblname.GetLength() > 0);
	if ( this->sdblib->sqllib->GetSql_Update(sql, this->tblname.GetText(), this->tbldef, data, where) < 0 )
		return -1;

	this->sdblib->sqllib->Begin();
	if ( this->sdblib->sqllib->Exec(sqlres, sql.GetText()) < 0
	  || this->sdblib->sqllib->GetStatus(sqlres) != SqlLib::RES_OK )
	{
		this->sdblib->sqllib->Rollback();
		return -1;
	}
	this->sdblib->sqllib->Commit();

	return 0;
}

int			DBTblLib::SelectData(SqlLib::Result & sqlres, const TCHAR * query)
{
	_ASSERT(this->sdblib != NULL && this->sdblib->sqllib != NULL);
	_ASSERT(this->tblname.GetLength() > 0);
	_ASSERT(query != NULL && query != TEXT('\0'));

	this->sdblib->sqllib->Begin();
	if ( this->sdblib->sqllib->Exec(sqlres, query) < 0
	  || this->sdblib->sqllib->GetStatus(sqlres) != SqlLib::RES_OK )
	{
		this->sdblib->sqllib->Rollback();
		return -1;
	}
	this->sdblib->sqllib->Commit();

	return 0;
}

int			DBTblLib::DeleteData(const TCHAR * where)
{
	SqlLib::VarBuffer	sql;
	SqlLib::Result		sqlres;

	_ASSERT(this->sdblib != NULL && this->sdblib->sqllib != NULL);
	_ASSERT(this->tblname.GetLength() > 0);
	sql.AddText(TEXT("DELETE FROM "));
	sql.AddText(this->tblname.GetText());
	if ( where != NULL )
	{
		sql.AddText(TEXT(" WHERE "));
		sql.AddText(where);
	}
	sql.AddNulChar();

	this->sdblib->sqllib->Begin();
	if ( this->sdblib->sqllib->Exec(sqlres, sql.GetText()) < 0
	  || this->sdblib->sqllib->GetStatus(sqlres) != SqlLib::RES_OK )
	{
		this->sdblib->sqllib->Rollback();
		return -1;
	}
	this->sdblib->sqllib->Commit();

	return 0;
}

int		GetSystemTime(SqlLib::TimeStamp & ts)
{
#if	defined(WIN32)
	SYSTEMTIME	wst;

	::GetSystemTime(&wst);
	ts.year   = (unsigned short)wst.wYear;
	ts.month  = (unsigned char )wst.wMonth;
	ts.day    = (unsigned char )wst.wDay;
	ts.hour   = (unsigned char )wst.wHour;
	ts.minute = (unsigned char )wst.wMinute;
	ts.second = (unsigned char )wst.wSecond;
#else
	time_t		tmbuf;
	struct tm	*gtm;

	time(&tmbuf);
	gtm = gmtime(&tmbuf);
	ts.year   = (unsigned short)gtm->tm_year + 1900;
	ts.month  = (unsigned char )gtm->tm_mon + 1;
	ts.day    = (unsigned char )gtm->tm_mday;
	ts.hour   = (unsigned char )gtm->tm_hour;
	ts.minute = (unsigned char )gtm->tm_min;
	ts.second = (unsigned char )gtm->tm_sec;
#endif
	return 0;
}

int		GetLocalTime(SqlLib::TimeStamp & lt, const SqlLib::TimeStamp & st)
{
#if	defined(WIN32)
	SYSTEMTIME	wst, wlt;

	wst.wYear         = st.year;
	wst.wMonth        = st.month;
	wst.wDayOfWeek    = 0;
	wst.wDay          = st.day;
	wst.wHour         = st.hour;
	wst.wMinute       = st.minute;
	wst.wSecond       = st.second;
	wst.wMilliseconds = 0;
	if ( ::SystemTimeToTzSpecificLocalTime(NULL, &wst, &wlt) == 0 )
		return -1;
	lt.year   = (unsigned short)wlt.wYear;
	lt.month  = (unsigned char )wlt.wMonth;
	lt.day    = (unsigned char )wlt.wDay;
	lt.hour   = (unsigned char )wlt.wHour;
	lt.minute = (unsigned char )wlt.wMinute;
	lt.second = (unsigned char )wlt.wSecond;
#else
	time_t	tmbuf;
	struct tm	gtm, *ltm;

	gtm.tm_year  = st.year - 1900;
	gtm.tm_mon   = st.month - 1;
	gtm.tm_mday  = st.day;
	gtm.tm_hour  = st.hour;
	gtm.tm_min   = st.minute;
	gtm.tm_sec   = st.second;
	gtm.tm_wday  = 0;
	gtm.tm_yday  = 0;
	gtm.tm_isdst = 0;
	tmbuf = mktime(&gtm);
	ltm = localtime(&tmbuf);
	lt.year   = (unsigned short)ltm->tm_year + 1900;
	lt.month  = (unsigned char )ltm->tm_mon + 1;
	lt.day    = (unsigned char )ltm->tm_mday;
	lt.hour   = (unsigned char )ltm->tm_hour;
	lt.minute = (unsigned char )ltm->tm_min;
	lt.second = (unsigned char )ltm->tm_sec;
#endif
	return 0;
}

int		CompTimeStamp(const SqlLib::TimeStamp & t1, const SqlLib::TimeStamp & t2)
{
	int		ret;

	if ( (ret = ((int)t1.year - (int)t2.year)) != 0 )
		return ret;
	if ( (ret = ((int)t1.month - (int)t2.month)) != 0 )
		return ret;
	if ( (ret = ((int)t1.day - (int)t2.day)) != 0 )
		return ret;
	if ( (ret = ((int)t1.hour - (int)t2.hour)) != 0 )
		return ret;
	if ( (ret = ((int)t1.minute - (int)t2.minute)) != 0 )
		return ret;
	if ( (ret = ((int)t1.second - (int)t2.second)) != 0 )
		return ret;
	return 0;
}
int		OffsetTimeStamp(SqlLib::TimeStamp & ts, const SqlLib::TimeStamp & src, int sec)
{
#if	defined(WIN32)
	SYSTEMTIME		wst;
	FILETIME		ft;	// 100nano sec
	ULARGE_INTEGER	tc;

	wst.wYear         = src.year;
	wst.wMonth        = src.month;
	wst.wDayOfWeek    = 0;
	wst.wDay          = src.day;
	wst.wHour         = src.hour;
	wst.wMinute       = src.minute;
	wst.wSecond       = src.second;
	wst.wMilliseconds = 0;
	if ( ::SystemTimeToFileTime(&wst, &ft) == 0 )
		return -1;
	tc.LowPart  = ft.dwLowDateTime;
	tc.HighPart = ft.dwHighDateTime;
	tc.QuadPart = (INT64)tc.QuadPart + (INT64)sec * I64(10000000);
	ft.dwLowDateTime  = tc.LowPart;
	ft.dwHighDateTime = tc.HighPart;
	if ( ::FileTimeToSystemTime(&ft, &wst) == 0 )
		return -1;
	ts.year   = (unsigned short)wst.wYear;
	ts.month  = (unsigned char )wst.wMonth;
	ts.day    = (unsigned char )wst.wDay;
	ts.hour   = (unsigned char )wst.wHour;
	ts.minute = (unsigned char )wst.wMinute;
	ts.second = (unsigned char )wst.wSecond;
#else
	time_t		tmbuf;
	struct tm	gtm, *tmp;

	gtm.tm_year  = src.year - 1900;
	gtm.tm_mon   = src.month - 1;
	gtm.tm_mday  = src.day;
	gtm.tm_hour  = src.hour;
	gtm.tm_min   = src.minute;
	gtm.tm_sec   = src.second;
	gtm.tm_wday  = 0;
	gtm.tm_yday  = 0;
	gtm.tm_isdst = 0;
	tmbuf = mktime(&gtm);
	tmbuf = (time_t)(Int64)tmbuf + (Int64)sec;
	tmp = gmtime(&tmbuf);
	ts.year   = (unsigned short)tmp->tm_year + 1900;
	ts.month  = (unsigned char )tmp->tm_mon + 1;
	ts.day    = (unsigned char )tmp->tm_mday;
	ts.hour   = (unsigned char )tmp->tm_hour;
	ts.minute = (unsigned char )tmp->tm_min;
	ts.second = (unsigned char )tmp->tm_sec;
#endif
	return 0;
}

//#include	<time.h>
int		DiffTimeStamp(const SqlLib::TimeStamp & t1, const SqlLib::TimeStamp & t2)
{
	const SqlLib::TimeStamp *	ts[2] = {&t1, &t2};
	int							i;
#if	 defined(WIN32)
	SYSTEMTIME					wst;
	FILETIME					ft;
	ULARGE_INTEGER				tc[2];

	for ( i = 0; i < 2; i++ )
	{
		wst.wYear         = ts[i]->year;
		wst.wMonth        = ts[i]->month;
		wst.wDayOfWeek    = 0;
		wst.wDay          = ts[i]->day;
		wst.wHour         = ts[i]->hour;
		wst.wMinute       = ts[i]->minute;
		wst.wSecond       = ts[i]->second;
		wst.wMilliseconds = 0;
		if ( ::SystemTimeToFileTime(&wst, &ft) == 0 )
		{
#if	defined(_DEBUG)
			_ftprintf(stderr, TEXT("sdb::DiffTimeStamp() error\n"));
#endif
			return -1;
		}
		tc[i].LowPart  = ft.dwLowDateTime;
		tc[i].HighPart = ft.dwHighDateTime;
	}
	return (int)( ((sdb::Int64)tc[0].QuadPart - (sdb::Int64)tc[1].QuadPart) / I64(10000000));
#else
	time_t		tmbuf[2];
	struct tm	gtm;

	for ( i = 0; i < 2; i++ )
	{
		gtm.tm_year  = ts[i]->year - 1900;
		gtm.tm_mon   = ts[i]->month - 1;
		gtm.tm_mday  = ts[i]->day;
		gtm.tm_hour  = ts[i]->hour;
		gtm.tm_min   = ts[i]->minute;
		gtm.tm_sec   = ts[i]->second;
		gtm.tm_wday  = 0;
		gtm.tm_yday  = 0;
		gtm.tm_isdst = 0;
		tmbuf[i] = mktime(&gtm);
	}
	return (int)((sdb::Int64)tmbuf[0] - (sdb::Int64)tmbuf[1]);
#endif
}

TCHAR *	Fmt(TCHAR * buffer, const TCHAR * form, ... )
{
	va_list	arg;

	va_start(arg, form);
	TForm::format(TForm::BUFFER_SIZE, buffer, form, arg);
	va_end(arg);
	return buffer;
}

int		CorrectValueName(SQStr & src)
{
	VarBuffer	wbuf;
	void		*p;
	int			ch;

	wbuf.Clear();
	for ( p = src.GetPtrTop(); p != NULL; p = src.GetPtrNext(p) )
	{
		if ( (ch = src.GetPtrChar(p)) <= 0 )
			break;
		if ( (ch >= TEXT('0') && ch <= TEXT('9') )
		  || (ch >= TEXT('A') && ch <= TEXT('Z') )
		  || (ch >= TEXT('a') && ch <= TEXT('z') )
		  ||  ch == TEXT('_')
		)
		{
			wbuf.AddChar(ch);
		} else
		{
			TCHAR	tmp[80];

			_stprintf(tmp, TEXT("%02X"), ch);
			wbuf.Add((unsigned)_tcslen(tmp), tmp);
		}
	}
	wbuf.AddChar(0);
	src.SetText(wbuf.GetText());
	return 0;
}

};// end of namespace sdb;


SdbLib::SdbLib()
{
	this->scrlib = new ScrLib;
	_ASSERT(this->scrlib != NULL);

	this->conn_info.host   = SDB_DEFAULT_HOST;
	this->conn_info.port   = SDB_DEFAULT_PORT;
	this->conn_info.dbname = SDB_DEFAULT_DBNAME;
	this->conn_info.user   = SDB_DEFAULT_USER;
	this->conn_info.pass   = SDB_DEFAULT_PASS;
	this->conn_info.option = NULL;

	this->sqllib = new SqlLib;
	_ASSERT(this->sqllib != NULL);

	this->dbdev.sdblib = this;

	return;
}

SdbLib::~SdbLib()
{
	if ( this->dev_list.size() > 0 )
		this->dev_clear();

	if ( this->sqllib != NULL )
	{
		if ( this->sqllib->IsConnect() )
			this->DBClose();
		delete this->sqllib;
	}

	if ( this->scrlib != NULL )
		delete this->scrlib;

	return;
}


void	SdbLib::dev_push_back(SdbDev * dev)
{
	this->dev_list.push_back(dev);
}

void	SdbLib::dev_clear(void)
{
	Dev_list::iterator	it;
	SdbDev				*dev;

	for ( it = this->dev_list.begin(); it != this->dev_list.end(); it++ )
	{
		dev = *it;
		if ( dev != NULL )
			delete dev;
	}
	this->dev_list.clear();
}

SdbDev	*SdbLib::dev_search(const TCHAR * desc)
{
	Dev_list::iterator	it;
	SdbDev				*dev;

	for ( it = this->dev_list.begin(); it != this->dev_list.end(); it++ )
	{
		dev = *it;
		if ( dev != NULL )
		{
			if ( SQStr::StrCmpi(dev->desc.GetText(), desc) == 0 )
				return dev;
		}
	}
	return NULL;
}

SdbDev *		SdbLib::CreateSdbDev(const TCHAR * desc)
{
	SdbDev	*dev;

	if ( (dev = new SdbDev) == NULL )
		return NULL;
	dev->dbdev = &this->dbdev;
	_ASSERT(dev->dbsmt != NULL);
	dev->dbsmt->sdblib = this;
	if ( desc != NULL )
		dev->desc.SetText(desc);

	return dev;
}

int		SdbLib::DBOpen(void)
{
	DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"SdbLib::DBOpen()\n"));
	if ( this->sqllib->IsConnect() )
		return -1;
	if ( this->sqllib->SetDBType(SqlLib::DBTYPE_PGLIB) < 0 )
		return -1;
	if ( this->sqllib->Connect(&this->conn_info) < 0 || this->sqllib->IsConnect() == FALSE )
		return -1;
	return 0;
}

int		SdbLib::DBClose(void)
{
	DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"SdbLib::DBClose()\n"));
	if ( this->sqllib->IsConnect() == FALSE )
		return 0;
	return this->sqllib->Finish();
}

int			SdbLib::GenerateSmtTblName(SQStr & tblname, sdb::DBDevData & devdata)
{
	TCHAR	*tmp;

	if ( (tmp = new TCHAR [8 + devdata.dev_model.GetLength() + 16]) == NULL )
		return -1;
	_stprintf(tmp, TEXT("smt_%s_id%d"), devdata.dev_model.GetDisplayText(), devdata.dev_id);
	tblname.SetText(tmp);
	sdb::CorrectValueName(tblname);
	delete [] tmp;

	return 0;
}

//	needs devdata.dev_model, dev_serial
sdb::DeviceOption *	SdbLib::GetDeviceOption(sdb::DBDevData & devdata)
{
	DevOptList::iterator	it;

	for ( it = this->devoptList.begin(); it != this->devoptList.end(); it++ )
	{
		sdb::DeviceOption *		devopt;

		devopt = *it;
		if ( SQStr::StrCmp(devopt->model.GetText() , devdata.dev_model.GetText() ) == 0 
		  && SQStr::StrCmp(devopt->serial.GetText(), devdata.dev_serial.GetText()) == 0 
		)
		{
			return devopt;
		}
	}
	return NULL;
}

