//////////////////////////////////////////////////////////////////////////
//	"sdbwatch.cpp"
//------------------------------------------------------------------------
//	$Id: sdbwatch.cpp 137 2006-04-25 02:42:37Z taro $
//////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#	include "config.h"
#	undef	HAVE_CONFIG_H
#endif
#if	defined(WIN32)	// libpq*.h  winsock.hCN[hwinsock2.hCN[h
# include	<winsock2.h>
#endif
#include	"sdbwatch.h"
#include	"../lib/smtplib.h"
#include	"../lib/reganalys.h"
#include	"../lib/getopt.h"
#include	"../lib/tform.h"
#if	defined(WIN32)
# include	"../win32/devctl.h"
#elif	defined(LINUX)
# include	"../linux/devctl.h"
# include	<ctype.h>
# include	<assert.h>
#endif
#include	<stdlib.h>
#include	<stdio.h>
#if defined(WIN32)
# include	<crtdbg.h>
#endif

#if	defined(_DEBUG) && defined(WIN32)
#	define	DBGREPORT(param_)	_CrtDbgReport param_
#elif	defined(DEBUG) && defined(LINUX)
#	define	_ASSERT	assert
#	define	DBGREPORT(param_)	DbgReport param_
#	define	_CRT_WARN			0
	extern int		DbgReport(int type, const char *filename, int linenumber, const char *moduleName, const char *format, ... );
#else
#	define	DBGREPORT(param_)
# ifndef	_ASSERT
#		define	_ASSERT
# endif
#endif

SdbWatch	sdbWatch;

enum	CmdOption
{
	OPT_DB_HOST       ,
	OPT_DB_NAME       ,
	OPT_DB_PASS       ,
	OPT_DB_PORT       ,
	OPT_DB_USER       ,
	OPT_DEFFILE       ,
	OPT_DEV           ,
	OPT_DISABLE_TITLE ,
	OPT_FAST          ,
	OPT_GROUPNAME     ,
	OPT_HELP          ,
	OPT_HOSTNAME      ,
	OPT_LONG          ,
	OPT_MODEL         ,
	OPT_OFFLINE_MARGIN,
	OPT_PREV_TIME     ,
	OPT_SERIAL        ,
	OPT_SENDMAIL      ,
};


static void	help(GetOpt & getopt);

int	_tmain(int argc, const TCHAR ** argv)
{
	int		retcode = EXIT_SUCCESS;

#if	0 // DEBUG
	{
		SqlLib::TimeStamp	ts, ts1, ts2;
		SQStr				str;

		sdb::GetSystemTime(ts);
		SdbWatch::GetDayTime(ts, str);
		_ftprintf(stdout, TEXT("Date: %s\n"), str.GetDisplayText());

		sdb::GetSystemTime(ts1);
		sdb::OffsetTimeStamp(ts2, ts1, -24 * 60 * 60);
		_ftprintf(stdout, TEXT("ts1 %04d-%02d-%02d %02d:%02d:%02d\n")
		  , ts1.year, ts1.month, ts1.day, ts1.hour, ts1.minute, ts1.second);
		_ftprintf(stdout, TEXT("ts2 %04d-%02d-%02d %02d:%02d:%02d\n")
		  , ts2.year, ts2.month, ts2.day, ts2.hour, ts2.minute, ts2.second);
	}
#endif
	if ( sdbWatch.Init() < 0 )
	{
		sdbWatch.PrintTitle();
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"sdbWatch.Init() error\n"));
		return EXIT_FAILURE;
	}
	if ( sdbWatch.CommandLine(argc, argv) < 0 )
		retcode = EXIT_FAILURE;
#if	defined(_DEBUG) && defined(WIN32)
	{
		TCHAR	tmp[1024];

		_tprintf(TEXT("*** EXIT ***\n"));
		_getts(tmp);
	}
#endif
	return retcode;
}

static void	help(GetOpt & getopt)
{
	int			i;

	static const TCHAR *msg1[] =
	{
	TEXT("usage: sdbwatch [<command>] [<option>...]"),
	TEXT("  <command>"),
	TEXT("    watch    .. Watch S.M.A.R.T. database"),
	TEXT("  <option>"),
	};
	for ( i = 0; i < sizeof(msg1)/sizeof(*msg1); i++)
		_putts(msg1[i]);

	getopt.coltbl[GetOpt::COL_1] =  2;
	getopt.coltbl[GetOpt::COL_2] =  4;
	getopt.coltbl[GetOpt::COL_3] = 28;
	getopt.coltbl[GetOpt::COL_4] = 31;

	static const GetOpt::HelpItem	opt1[] =
	{
	{ NULL, OPT_DB_HOST       , TEXT("<hostname>"), TEXT("Database hostname    (Default ") SDB_DEFAULT_HOST   TEXT(")")},
	{ NULL, OPT_DB_PASS       , TEXT("<password>"), TEXT("Database password    (Default ") SDB_DEFAULT_PASS   TEXT(")")},
	{ NULL, OPT_DB_PORT       , TEXT("<port#>"   ), TEXT("Database port-number (Default ") SDB_DEFAULT_PORT   TEXT(")")},
	{ NULL, OPT_DB_USER       , TEXT("<username>"), TEXT("Database username    (Default ") SDB_DEFAULT_USER   TEXT(")")},
	{ NULL, OPT_DB_NAME       , TEXT("<name>"    ), TEXT("Database name        (Default ") SDB_DEFAULT_DBNAME TEXT(")")},
	{ NULL, OPT_DEFFILE       , TEXT("<path>"    ), TEXT("Deffile path"        )},
	{ NULL, OPT_DEV           , TEXT("<device>"  ), TEXT("Device"              )},
	{ NULL, OPT_FAST          , TEXT(""          ), TEXT("Fast check"          )},
	{ NULL, OPT_GROUPNAME     , TEXT("<name>"    ), TEXT("Groupname"           )},
	{ NULL, OPT_HOSTNAME      , TEXT("<hostname>"), TEXT("Local PC hostname"   )},
	{ NULL, OPT_LONG          , TEXT(""          ), TEXT("Long check"          )},
	{ NULL, OPT_MODEL         , TEXT("<model>"   ), TEXT("Device model name  (effective --dev")},
	{ NULL, OPT_OFFLINE_MARGIN, TEXT("<sec>"     ), TEXT("Offline margine time")},
	{ NULL, OPT_PREV_TIME     , TEXT("<sec>"     ), TEXT("Previous time")},
	{ NULL, OPT_SERIAL        , TEXT("<serial>"  ), TEXT("Device Serial name (effective --dev")},
	{ NULL, OPT_SENDMAIL      , TEXT(""          ), TEXT("Send mail")},
	};
	getopt.Help(sizeof(opt1)/sizeof(*opt1), opt1);

	return;
}


//////////////////////////////////////////////////////////////////////////
//	SdbWatch
//////////////////////////////////////////////////////////////////////////

SdbWatch::SdbWatch()
{
	this->sdblib = new SdbLib;	_ASSERT(this->sdblib != NULL);
	this->dbwat  = sdb::DBWatLib::Create(this->sdblib); _ASSERT(this->dbwat != NULL);

	this->print_title = FALSE;

	this->opt.checkmode = CHECK_FAST;
	this->opt.offline_margin_sec = 2 * 60 * 60;	// 2 hour
	this->opt.prev_time_sec      = 60 * 60;	// 1 hour
	this->opt.disable_title      = FALSE;

//	this->curWatchDev = NULL;

	this->watnum = 0;
	this->watlist = NULL;

	return;
}

SdbWatch::~SdbWatch()
{
	this->ClearWatchDevList();

	if ( this->watlist != NULL )
		delete [] this->watlist;

	if ( this->dbwat != NULL )
		delete this->dbwat;

	if ( this->sdblib != NULL )
		delete this->sdblib;
	return;
}

void	SdbWatch::PrintTitle(void)
{
	if ( this->print_title == FALSE && this->opt.disable_title == FALSE )
	{
		_putts(SDBWATCH_TITLE);
		this->print_title = TRUE;
	}
	return;
}

int		SdbWatch::Init(void)
{
	int		retcode;

	try
	{
		if ( this->sdblib  == NULL )
			return -1;

#if	! defined(WIN32)
		tzset();
#endif

#if	defined(WIN32)
		// NpX̎擾
		{
			TCHAR	buf[1024];
			int		len;

			if ( (len = ::GetModuleFileName(NULL, buf, sizeof(buf)/sizeof(*buf))) > 0 )
				this->progPath.SetText(buf);
			this->progDir.GetDnam(this->progPath);
		}
#elif	defined(LINUX)
		{
			this->progDir.SetText(SDB_LOADDEF_DIR);
		}
#endif

		// deft@C
		{
#if	defined(SDB_DEFFILENAME)
			this->defFile.SetText(this->progDir);
			this->defFile.AddPath(SDB_DEFFILENAME);
#else
			this->defFile.SetText(this->progPath);
			this->defFile.GetFn(path);
			this->defFile.AddText(TEXT(".def"));
#endif
		}

		retcode = 0;
	} catch(int ret) { retcode = ret; };

	return retcode;
}

int		SdbWatch::LoadDef(void)
{
	if ( this->defFile.GetLength() == 0 )
		return 0;	// no def file

	if ( this->sdblib->LoadScr(this->defFile.GetText()) < 0 )
	{
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"sdblib->LoadScr(%s) error\n", this->defFile.GetDisplayText()));
	//	_ftprintf(stderr, TEXT("Load %s error!\n"), this->defFile.GetDisplayText());
		return -1;
	}

	return 0;
}

void		SdbWatch::ClearWatchDevList(void)
{
	WatchDevList::iterator	it;

//	this->curWatchDev = NULL;
	for ( it = this->watchDevList.begin(); it != this->watchDevList.end(); it++ )
	{
		WatchDev	*watdev;

		watdev = *it;
		if ( watdev != NULL )
			delete watdev;
	}
	this->watchDevList.clear();
}

int		SdbWatch::CommandLine(int argc, const TCHAR ** argv)
{
	GetOpt							getopt;
	enum	ArgError
	{
		ARG_SUCCESS     ,
		ARG_HELP        ,
		ARG_ERROR_PARAM ,
		ARG_ERROR_GETOPT,
		ARG_ERROR       ,
	};
	ArgError		argError;
	SQStr			errmsg;
	CtlCmd			cmd = CMD_HELP;
	SdbLib			*sdblib = this->sdblib;
	SdbDev			*sdbdev = NULL;
	int				retcode = 0;

#if	0 // DEBUG
	{
		SqlLib::TimeStamp	ts, ts1, ts2;
		SQStr				str;

		sdb::GetSystemTime(ts);
		SdbWatch::GetDayTime(ts, str);
		_ftprintf(stdout, TEXT("Date: %s\n"), str.GetDisplayText());

		sdb::GetSystemTime(ts1);
		sdb::OffsetTimeStamp(ts2, ts1, -24 * 60 * 60);
		_ftprintf(stdout, TEXT("ts1 %04d-%02d-%02d %02d:%02d:%02d\n")
		  , ts1.year, ts1.month, ts1.day, ts1.hour, ts1.minute, ts1.second);
		_ftprintf(stdout, TEXT("ts2 %04d-%02d-%02d %02d:%02d:%02d\n")
		  , ts2.year, ts2.month, ts2.day, ts2.hour, ts2.minute, ts2.second);
	}
#endif

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"_tmail(argc=%d)\n",argc));
	{
		static GetOpt::AddItemParam	tbl[] =
		{
			{ OPT_DB_HOST           , TEXT("db_host")       , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_DB_PASS           , TEXT("db_pass")       , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_DB_PORT           , TEXT("db_port")       , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_DB_USER           , TEXT("db_user")       , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_DB_NAME           , TEXT("db_name")       , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_DEFFILE           , TEXT("deffile")       , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_DEV               , TEXT("dev")           , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_DISABLE_TITLE     , TEXT("disable_title") , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_FAST              , TEXT("fast")          , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_GROUPNAME         , TEXT("groupname")     , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_HELP              , TEXT("help")          , NULL, GetOpt::VAR_FLAG  , FALSE, { NULL,0,0} },
			{ OPT_HOSTNAME          , TEXT("hostname")      , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_LONG              , TEXT("long")          , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_MODEL             , TEXT("model")         , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_OFFLINE_MARGIN    , TEXT("offline_margin"), NULL, GetOpt::VAR_INT   , FALSE, { NULL,0,0} },
			{ OPT_PREV_TIME         , TEXT("prev_time")     , NULL, GetOpt::VAR_INT   , FALSE, { NULL,0,0} },
			{ OPT_SERIAL            , TEXT("serial")        , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_SENDMAIL          , TEXT("sendmail")      , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
		};

		if ( getopt.AddItem(sizeof(tbl)/sizeof(*tbl), tbl) < 0 )
		{
			this->PrintTitle();
			_ftprintf(stderr, TEXT("Getopt AddItem Error!\n"));
			return -1;
		}
	}

	if ( this->Init() < 0 )
	{
		this->PrintTitle();
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"sdbWatch.Init() error\n"));
		return -1;
	}

	try
	{
		int			an;

		an = 1;
		getopt.SetCurPos(an);
		while ( getopt.GetCurPos() < argc )
		{
			GetOpt::Item	*item;
			const char		*s;

			if ( (item = getopt.ParseArg(an, argc, argv)) == NULL )
			{
				if ( (s = getopt.last_param) != NULL )
				{
					if ( _tcsicmp(s, TEXT("help")) == 0 )
						throw ARG_HELP;
					else if ( _tcsicmp(s, TEXT("watch")) == 0 )
						cmd = CMD_WATCH;
					else
						throw ARG_ERROR_GETOPT;
				} else if ( getopt.err_option != NULL || getopt.err_param != NULL )
				{
					throw ARG_ERROR_GETOPT;
				} // end of Arg
			} else
			{
				switch ( item->usertype )
				{
				case OPT_DB_HOST:
					sdblib->dbinfo.host.SetText(item->var.text);
					break;

				case OPT_DB_PASS:
					sdblib->dbinfo.pass.SetText(item->var.text);
					break;

				case OPT_DB_PORT:
					sdblib->dbinfo.port.SetText(item->var.text);
					break;

				case OPT_DB_USER:
					sdblib->dbinfo.user.SetText(item->var.text);
					break;

				case OPT_DB_NAME:
					sdblib->dbinfo.dbname.SetText(item->var.text);
					break;

				case OPT_DEFFILE:
					this->defFile.SetText(item->var.text);
					break;

				case OPT_DEV:
				{
					const TCHAR		*desc;
					TCHAR			tmp[1024];

					if ( (desc = item->var.text) == NULL )
						throw ARG_ERROR_PARAM;
					if ( sdblib->dev_search(item->var.text) != NULL )
					{
						_stprintf(tmp, TEXT("Device alreay use! (%)\n"), desc);
						errmsg.SetText(tmp);
						throw ARG_ERROR;
					}
					if ( (sdbdev = sdblib->CreateSdbDev(desc)) == NULL )
						throw ARG_ERROR;
					sdblib->dev_push_back(sdbdev);
					break;
				}

				case OPT_DISABLE_TITLE:
					this->opt.disable_title = item->var.flag;
					break;

				case OPT_FAST:
					if ( item->var.flag )
						this->opt.checkmode = SdbWatch::CHECK_FAST;
					break;

				case OPT_GROUPNAME:
					if ( sdbdev == NULL )
						throw ARG_ERROR_PARAM;
					sdbdev->groupname.SetText(item->var.text);
					break;

				case OPT_HELP:
					throw ARG_HELP;
					break;

				case OPT_HOSTNAME:
					if ( sdbdev == NULL )
						throw ARG_ERROR_PARAM;
					sdbdev->hostname.SetText(item->var.text);
					break;

				case OPT_LONG:
					if ( item->var.flag )
						this->opt.checkmode = SdbWatch::CHECK_LONG;
					break;

				case OPT_MODEL:
					if ( sdbdev == NULL )
						throw ARG_ERROR_PARAM;
					sdbdev->model.SetText(item->var.text);
					break;

				case OPT_OFFLINE_MARGIN:
					this->opt.offline_margin_sec = item->var.n;
					break;

				case OPT_PREV_TIME:
					this->opt.prev_time_sec = item->var.n;
					break;

				case OPT_SERIAL:
					if ( sdbdev == NULL )
						throw ARG_ERROR_PARAM;
					sdbdev->serial.SetText(item->var.text);
					break;

				case OPT_SENDMAIL:
					this->opt.sendmail = item->var.flag;
					break;

				default:
					_ASSERT(0);
					break;
				} // end of switch
			}
		} // end of while
		argError = ARG_SUCCESS;
	} catch(ArgError code) { argError = code; }

	this->PrintTitle();
	if ( argError != ARG_SUCCESS )
	{
		switch ( argError )
		{
		case ARG_HELP:
			help(getopt);
			break;

		case ARG_ERROR_PARAM:
			_ftprintf(stderr, TEXT("Parametr error! (%s)\n"), getopt.GetLastPos() >= 0  ? argv[getopt.GetLastPos()] : TEXT("") );
			break;

		case ARG_ERROR_GETOPT:
			_ftprintf(stderr, TEXT("Option parameter error!\n"));
			if ( getopt.err_option != NULL )
				_ftprintf(stderr, TEXT("  option   =\"%s\"\n"), getopt.err_option);
			if ( getopt.err_param != NULL )
				_ftprintf(stderr, TEXT("  parameter=\"%s\"\n"), getopt.err_param);
			break;

		case ARG_ERROR:
		default:
			if ( errmsg.GetLength() == 0 )
				errmsg.SetText(TEXT("Internal error!\n"));
			_fputts(errmsg.GetDisplayText(), stderr);
			break;
		}
		return -1;
	}
	if ( this->LoadDef() < 0 )
		return -1;

	{
		_ASSERT(sdblib != NULL);
		sdblib->conn_info.host   = sdblib->dbinfo.host.GetLength()   > 0 ? sdblib->dbinfo.host.GetText()  : (TCHAR *)sdb::db_host;
		sdblib->conn_info.port   = sdblib->dbinfo.port.GetLength()   > 0 ? sdblib->dbinfo.port.GetText()  : (TCHAR *)sdb::db_port;
		sdblib->conn_info.user   = sdblib->dbinfo.user.GetLength()   > 0 ? sdblib->dbinfo.user.GetText()  : (TCHAR *)sdb::db_user;
		sdblib->conn_info.pass   = sdblib->dbinfo.pass.GetLength()   > 0 ? sdblib->dbinfo.pass.GetText()  : (TCHAR *)sdb::db_pass;
		sdblib->conn_info.dbname = sdblib->dbinfo.dbname.GetLength() > 0 ? sdblib->dbinfo.dbname.GetText(): (TCHAR *)sdb::db_name;
		if ( sdblib->DBOpen() < 0 )
		{
			_ftprintf(stderr, TEXT("DBOpen error!\n"));
		}
	}
	if ( this->sdblib->dev_list.size() > 0 )
		this->LocalDevice();

	switch ( cmd )
	{
	case CMD_WATCH:
		this->Watch();
		break;
	}

	sdblib->DBClose();

	return retcode;
}


//////////////////////////////////////////////////////////////////////////
//	Check Local Device

int		SdbWatch::LocalDevice(void)
{
	SdbLib::Dev_list::iterator	it;
	SdbDev						*sdbdev;

//	printf(TEXT("DEBUG: LocalDevice()\n"));
	for ( it = this->sdblib->dev_list.begin(); it != this->sdblib->dev_list.end(); it++ )
	{
		sdbdev = *it;
		_ASSERT(sdbdev != NULL);
		if ( sdbdev->model.GetLength() > 0 && sdbdev->serial.GetLength() > 0 )
		{
		} else if ( sdbdev->desc.GetLength() > 0 )
		{
			this->LocalDevice(sdbdev);
		}
	}
	return 0;
}

int		SdbWatch::LocalDevice(SdbDev * dev)
{
	unsigned	rwmode;

	_ASSERT(dev != NULL);
	rwmode = DevCtl::RW_READ | DevCtl::RW_WRITE + DevCtl::RW_SHARE_READ + DevCtl::RW_SHARE_WRITE + DevCtl::RW_NOBUFFER;
	if ( dev->OpenDevice(rwmode) < 0 )
		return -1;
	if ( /*dev->GetSmart() == 0 &&*/ dev->GetDeviceData() == 0 )
	{
		
	}
	dev->CloseDevice();
	return 0;
}


//////////////////////////////////////////////////////////////////////////
//	Watch Smart Database

int		SdbWatch::Watch(void)
{
	SqlLib				*sqllib = NULL;
	SqlLib::VarBuffer	sql, where;
	SqlLib::Result		sqlres;
	int					xmax, ymax;

	_ASSERT(this->sdblib != NULL);
	if ( this->sdblib->DBIsConnect() == FALSE )
	{
		_ftprintf(stderr, TEXT("Offline database!\n"));
		return -1;
	}
	sqllib = this->sdblib->dbdev.GetSqlLib();
	_ASSERT(sqllib != NULL);
	{
		where.AddText(sdb::dev_tbl[sdb::DEV_F_DISABLE].fname );
		where.AddText(TEXT("<>true AND "));
		where.AddText(sdb::dev_tbl[sdb::DEV_SMT_TBLNAME].fname );
		where.AddText(TEXT("<>''"));
		if ( this->sdblib->dev_list.size() > 0 )
		{
			SdbLib::Dev_list::iterator	it;
			SdbDev						*sdbdev;
			int							dev_num = 0;

			for ( it = this->sdblib->dev_list.begin(); it != this->sdblib->dev_list.end(); it++ )
			{
				sdbdev = *it;
				_ASSERT(sdbdev != NULL);
				if ( sdbdev->model.GetLength() > 0 && sdbdev->serial.GetLength() > 0 )
				{	// foCXw
					if ( dev_num == 0 )
						where.AddText(TEXT(" AND (("));
					else
						where.AddText(TEXT(" OR ("));

					where.AddText(sdb::dev_tbl[sdb::DEV_MODEL ].fname );
					where.AddText(TEXT("="));
					SqlLib::EscapeString(where, sdbdev->model.GetText(), TEXT('\''));

					where.AddText(TEXT(" AND "));

					where.AddText(sdb::dev_tbl[sdb::DEV_SERIAL].fname );
					where.AddText(TEXT("="));
					SqlLib::EscapeString(where, sdbdev->serial.GetText(), TEXT('\''));

					where.AddText(TEXT(")"));
					dev_num++;
				}
			}
			if ( dev_num > 0 )
				where.AddText(TEXT(")"));
		}
		where.AddChar(0);
	//	printf("DEBUG: where = %s\n", where.GetText());	// DEBUG
	}
	sql.AddText(TEXT("SELECT * FROM "));
	sql.AddText(this->sdblib->dbdev.tblname.GetText());
	sql.AddText(TEXT(" WHERE "));
//	sql.AddText(sdb::dev_tbl[sdb::DEV_F_DISABLE].fname );
//	sql.AddText(TEXT("<>true AND "));
//	sql.AddText(sdb::dev_tbl[sdb::DEV_SMT_TBLNAME].fname );
//	sql.AddText(TEXT("<>''"));
	sql.AddText(where.GetText());
	sql.AddText(TEXT(" ORDER BY "));
	sql.AddText(sdb::dev_tbl[sdb::DEV_GROUPNAME].fname );
	sql.AddText(TEXT(","));
	sql.AddText(sdb::dev_tbl[sdb::DEV_HOSTNAME].fname );
	sql.AddText(TEXT(","));
	sql.AddText(sdb::dev_tbl[sdb::DEV_MODEL].fname );
	sql.AddChar(0);

	if ( this->sdblib->dbdev.SelectData(sqlres, sql.GetText()) < 0 )
	{
		return -1;
	}
	ymax = sqllib->GetTuples(sqlres);
	xmax = sqllib->GetFields(sqlres);
	if ( xmax > 0 && ymax > 0 )
	{
		int		y;

		for ( y = 0; y < ymax; y++ )
		{
			sdb::DBDevData		devdata;
			SqlLib::FieldsTbl	fdata;
			int					x;

			fdata.Clear();
			for ( x = 0; x < xmax; x++ )
			{
				const TCHAR			*fname;
				SqlLib::DataValue	val;

				if ( (fname  = sqllib->GetFname(sqlres, x)) != NULL )
				{
					if ( sqllib->GetDataValue(sqlres, y, x, val) >= 0 )
						fdata.Add(fname, val);
				}
			}
			this->sdblib->dbdev.GetDevData(devdata, fdata);
			{
				WatchDev	*watdev;

				if ( (watdev = new WatchDev) == NULL )
					return -1;
				watdev->devdata = devdata;
				this->watchDevList.push_back(watdev);
			}
		}
	}
	sqllib->ClearResult(sqlres);

	if ( this->GetWatchConf() < 0 )
	{
		_ftprintf(stderr, TEXT("Watch table nothing.\n"));
		if ( this->watlist == NULL )
			this->watnum = 0;
	}

	{
		WatchDevList::iterator	it;

		for ( it = this->watchDevList.begin(); it != this->watchDevList.end(); it++ )
		{
			WatchDev	*watdev;

			watdev = *it;
			this->WatchDevice(*watdev);
		}
	}

	return 0;
}

int		SdbWatch::GetWatchConf(void)
{
	SqlLib				*sqllib = NULL;
	SqlLib::VarBuffer	sql;
	SqlLib::Result		sqlres;
	int					xmax, ymax;
	int					retcode = 0;

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"SdnWatch::GetWatchConf()\n"));
	_ASSERT(this->sdblib != NULL);
	if ( this->sdblib->DBIsConnect() == FALSE )
		return -1;
	_ASSERT(this->dbwat != NULL);
	sqllib = this->dbwat->GetSqlLib();
	_ASSERT(sqllib != NULL);

	sql.AddText(TEXT("SELECT * FROM "));
	sql.AddText(this->dbwat->tblname.GetText());
	sql.AddText(TEXT(" WHERE "));
	sql.AddText(sdb::wat_tbl[sdb::DEV_F_DISABLE].fname );
	sql.AddText(TEXT("<>true"));
	sql.AddText(TEXT(" ORDER BY "));
	sql.AddText(sdb::wat_tbl[sdb::WAT_PRIORITY].fname );
	sql.AddText(TEXT(","));
	sql.AddText(sdb::wat_tbl[sdb::WAT_ID].fname );
	sql.AddText(TEXT(","));
	sql.AddText(sdb::wat_tbl[sdb::WAT_ATTR_ID].fname );
	sql.AddChar(0);

	if ( this->dbwat->SelectData(sqlres, sql.GetText()) < 0 )
		return -1;

	ymax = sqllib->GetTuples(sqlres);
	xmax = sqllib->GetFields(sqlres);
	if ( xmax > 0 && ymax > 0 )
	{
		int		y;

		this->watnum = ymax;
		if ( this->watlist != NULL )
			delete [] this->watlist;
		if ( (this->watlist = new sdb::DBWatData[this->watnum]) == NULL )
		{
			retcode = 0;
		} else
		{
			for ( y = 0; y < ymax; y++ )
			{
				SqlLib::FieldsTbl	fdata;
				int					x;

				fdata.Clear();
				for ( x = 0; x < xmax; x++ )
				{
					const TCHAR			*fname;
					SqlLib::DataValue	val;

					if ( (fname  = sqllib->GetFname(sqlres, x)) != NULL )
					{
						if ( sqllib->GetDataValue(sqlres, y, x, val) >= 0 )
							fdata.Add(fname, val);
					}
				}
				this->dbwat->GetWatData(this->watlist[y], fdata);
			}
			retcode = 0;
		}
	}
	sqllib->ClearResult(sqlres);

#if	defined(_DEBUG)
	if ( retcode == 0 )
	{
		unsigned	y;

		for ( y = 0; y < this->watnum; y++ )
		{
			sdb::DBWatData	watdata;

			watdata = this->watlist[y];
			_tprintf(TEXT("%3d model=\"%s\", serial=\"%s\", attr_id=%d, type=%d. warn=%d, op=\"%s\", right=%s \n")
			  , y
			  , watdata.wat_model.GetDisplayText()
			  , watdata.wat_serial.GetDisplayText()
			  , (int)watdata.wat_attr_id
			  , (int)watdata.wat_type
			  , (int)watdata.wat_warn
			  , watdata.wat_op.GetDisplayText()
			  , watdata.wat_right.GetDisplayText()
			);
		}
	}
#endif

	return retcode;
}

#if	defined(_DEBUG)
void	dbg_print_time(const TCHAR * msg)
{
	SqlLib::TimeStamp	ts;

	sdb::GetSystemTime(ts);
	sdb::GetLocalTime(ts, ts);
	DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"%s : %04d-%02d-%02d %02d:%02d:%02d\n"
	  , msg
	  , (int)ts.year, (int)ts.month , (int)ts.day
	  , (int)ts.hour, (int)ts.minute, (int)ts.second
	));
	return;
}
#else
#	define	dbg_print_time
#endif

int		SdbWatch::WatchDevice(WatchDev & watdev)
{
	SqlLib				*sqllib = NULL;
	VarBuffer			sql, where;
	SqlLib::Result		sqlres;
	SqlLib::TimeStamp	tsnow, ts, lt;
	BOOL				online = TRUE;
	SdbDev				*sdbdev = NULL;
	int					retcode = 0;
	sdb::DBSmtData		*smtlist = NULL;
	AttrMap				attrMap;
	TCHAR				tmp[256];

	this->watmsg.Clear();
	this->WarnClear();
//	this->curWatchDev = &watdev;
	try
	{
		TCHAR	desc[80];

		_stprintf(desc, TEXT("dev_id=%d"), (int)watdev.devdata.dev_id);
		if ( (sdbdev = this->sdblib->CreateSdbDev(desc)) == NULL )
			throw -1;

		this->WatchMessage(TEXT("\n"));
		if ( watdev.devdata.dev_hostname.GetLength() > 0 )
		{
			this->WatchMessage(sdb::Fmt(tmp, TEXT("Hostname %s\n")
			  , watdev.devdata.dev_hostname.GetDisplayText()
			));
		}
		this->WatchMessage(sdb::Fmt(tmp, TEXT("%Model %s, Serial %s\n")
		  , watdev.devdata.dev_model.GetDisplayText()
		  , watdev.devdata.dev_serial.GetDisplayText()
		));
		sqllib = this->sdblib->dbdev.GetSqlLib();
		_ASSERT(sqllib != NULL);
		sdb::GetSystemTime(tsnow);
		if ( this->opt.offline_margin_sec < 0 )
			this->opt.offline_margin_sec = -(this->opt.offline_margin_sec);
		sdb::OffsetTimeStamp(ts, tsnow, -(this->opt.offline_margin_sec));	// 2 hour
		sdb::GetLocalTime(lt, watdev.devdata.dev_ts_last);
		this->WatchMessage(sdb::Fmt(tmp, TEXT("  Last update %04d-%02d-%02d %02d:%02d:%02d\n") 
			  , (int)lt.year, (int)lt.month , (int)lt.day
			  , (int)lt.hour, (int)lt.minute, (int)lt.second
		));
		
		if ( sdb::CompTimeStamp(watdev.devdata.dev_ts_last, ts) < 0 )
		{	// offline
			this->WatchMessage(TEXT("  off-line\n"));
			online = FALSE;
		}
		where.Printf(TEXT("%s=%d"), sdb::smt_tbl[sdb::SMT_DEV_ID].fname, (int)watdev.devdata.dev_id);

		switch ( this->opt.checkmode )
		{
		case CHECK_FAST:
		{
			SqlLib::DataValue	val;
			SqlLib::VarBuffer	vtmp;

			if ( online )
			{	// 24 ԑÕf[^擾
				sdb::OffsetTimeStamp(ts, tsnow, -24 * 60 * 60);	// 24 hour
				val.SetTimeStamp(ts);
			} else
				val.SetTimeStamp(watdev.devdata.dev_ts_last);
			vtmp.Clear();
			sqllib->SetSqlValue(vtmp, sdb::smt_tbl[sdb::SMT_TS], val);
			vtmp.AddChar(0);
			where.Printf(TEXT(" AND %s>=%s"), sdb::smt_tbl[sdb::SMT_TS].fname, vtmp.GetText());
			break;
		}
		case CHECK_LONG:
			break;
		default:
			_ASSERT(0);
			throw -1;
		}
		where.AddChar(0);
		sql.Printf(TEXT("SELECT * FROM %s WHERE %s ORDER BY %s,%s")
		  , watdev.devdata.dev_smt_tblname.GetText()
		  , where.GetText()
		  , sdb::smt_tbl[sdb::SMT_TS     ].fname
		  , sdb::smt_tbl[sdb::SMT_ATTR_ID].fname
		);
		sql.AddChar(0);
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"sql = %s\n", sql.GetText()));

		_ASSERT(sdbdev->dbsmt != NULL);
		dbg_print_time("Bgin Select");
		if ( sdbdev->dbsmt->SelectData(sqlres, sql.GetText()) < 0 )
			throw -1;
		dbg_print_time("End Select");

		dbg_print_time("Begin create SmtList");
		{
			int				xmax, ymax;
			int				x, y;
			const TCHAR		**fname_tbl = NULL;

			ymax = sqllib->GetTuples(sqlres);
			xmax = sqllib->GetFields(sqlres);
			DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"xmax = %d, ymax = %d\n", xmax, ymax));
			if ( xmax <= 0 || ymax <= 0 )
				throw -1;
			if ( (smtlist = new sdb::DBSmtData[ymax]) == NULL )
				throw -1;
			if ( (fname_tbl = new const TCHAR * [xmax]) == NULL )
				throw -1;
			for ( x = 0; x < xmax; x++ )
				fname_tbl[x] = sqllib->GetFname(sqlres, x);
			for ( y = 0; y < ymax; y++ )
			{
				sdbdev->dbsmt->GetSmtData(smtlist[y], sqlres, y, xmax, fname_tbl);
			}
			if ( fname_tbl != NULL )
				delete [] fname_tbl;
			if ( SdbWatch::GetAttributeNum(attrMap, ymax, smtlist) == 0 )
				throw -1;
		}
		dbg_print_time("End create SmtList");

		{
			unsigned	attr_id;

			PutPos(4     , TEXT("ID# S.M.A.R.T. Attribute"),
			       4 + 4+sdb::SMART_ATTRIBUTE_MAX_LENGHTH, TEXT("Val Wst Ths Flags"),
			       COLUMN_MAX - 1 - 6 - 1 - 9            , TEXT("RAW Value"),
			       COLUMN_MAX - 1 - 6                    , TEXT("Status"),
			       0);
			for ( attr_id = 0; attr_id < SDB_SMART_ATTR_ID_MAX; attr_id++ )
			{
				sdb::DBSmtData	smtdata;
				AttrData *		attrData;
				unsigned		wat_no;
				BOOL			attr_failer;

				attrData = &attrMap.id[attr_id];
				if ( attrData->count == 0 || attrData->smtlist == NULL )
					continue;	// skip
				if ( attr_id == sdb::auto_raw_word_id )
				{	// RAW Value16rbgl̏ꍇ́AIɌグ
					unsigned	i;
					sdb::Int64	prev, v, add_value;
					BOOL		flag;

					flag = TRUE;
					prev = I64(0);
					add_value = I64(0);
					for ( i = 0; i < attrData->count; i++ )
					{
						v = attrData->smtlist[i]->smt_raw_value;
						if ( v < prev )
						{
							if ( prev > SDB_RAW_WORD_MAX - SDB_RAW_WORD_MARGIN && v < SDB_RAW_WORD_MARGIN )
							{
								add_value += I64(65536);
							} else
							{
								flag = FALSE;
								break;
							}
						}
						prev = v;
					}
					if ( flag )
					{
						prev = add_value = I64(0);
						for ( i = 0; i < attrData->count; i++ )
						{
							v = attrData->smtlist[i]->smt_raw_value;
							if ( v < prev && prev > SDB_RAW_WORD_MAX - SDB_RAW_WORD_MARGIN && v < SDB_RAW_WORD_MARGIN )
								add_value += I64(65536);
							attrData->smtlist[i]->smt_raw_value += add_value;
							prev = v;
						}
					}
				}
				smtdata = *attrData->smtlist[attrData->count - 1];
				this->PrintSmartData(watdev.devdata, smtdata);
				attr_failer = smtdata.smt_value <= smtdata.smt_threshold ? TRUE : FALSE;
				for ( wat_no = 0; wat_no < this->watnum; wat_no++ )
				{
					sdb::DBWatData	watdata;
					sdb::Int64		left_value, right_value;

					watdata = this->watlist[wat_no];
					if ( watdata.wat_attr_id != attr_id )
						continue;	// Attribute-ID mismatch

					// fmF
					if ( watdata.wat_model.GetLength() == 0 )
					{	// fȂ
						if ( attrData->ww[watdata.wat_type].watdata.wat_attr_id > 0
						  && attrData->ww[watdata.wat_type].watdata.wat_model.GetLength() > 0
						)
						{	// ɏ΁AsȂ
							DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"watch rule %u, id %u skip check 1\n", watdata.wat_id, attr_id));
							continue;
						}
					} else if ( SQStr::StrCmp(watdev.devdata.dev_model.GetText(), watdata.wat_model.GetText(), watdata.wat_model.GetLength())== 0 )
					{	// fv
						if ( attrData->ww[watdata.wat_type].watdata.wat_attr_id > 0
						  && attrData->ww[watdata.wat_type].watdata.wat_model.GetLength() > 0
						  && SQStr::StrCmp(attrData->ww[watdata.wat_type].watdata.wat_model.GetText(), watdata.wat_model.GetText()) != 0 
						)
						{	// OƃfقȂ
							DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"watch rule %u, id %u skip: check 2\n", watdata.wat_id, attr_id));
							continue;
						}
						if ( watdata.wat_serial.GetLength() > 0 && SQStr::StrCmp(watdev.devdata.dev_serial.GetText(), watdata.wat_serial.GetText()) == 0 )
						{	// VAv
							if ( attrData->ww[watdata.wat_type].watdata.wat_attr_id > 0
							  && attrData->ww[watdata.wat_type].watdata.wat_serial.GetLength() > 0
							  && SQStr::StrCmp(attrData->ww[watdata.wat_type].watdata.wat_serial.GetText(), watdata.wat_serial.GetText()) != 0 
							)
							{	// OƃVAقȂ
								DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"watch rule %u, id %u skip: check 3\n", watdata.wat_id, attr_id));
								continue;
							}
						}
					} else
						continue;

					attrData->ww[watdata.wat_type].count++;
					attrData->ww[watdata.wat_type].watdata = watdata;

					// LEFT VALUE
					if ( this->GetWatValue(left_value, watdata.wat_type, *(attrData->smtlist[attrData->count-1])) <= 0 )
					{	// no left_value
						if ( watdata.wat_type == sdb::WATCHTYPE_OFF )
						{
							attr_failer = FALSE;	// ͖
							continue;
						}
						// Analys
						if ( watdata.wat_type == sdb::WATCHTYPE_RAW_ANALYS && this->opt.checkmode == CHECK_LONG )
						{
						//	this->WatchMessage(TEXT("DEBUG: RAW TEC\n"));
							right_value = _ttoi64(watdata.wat_right.GetText());
							this->AnalysRawValue(attr_id, right_value, *attrData, tsnow);
						}
					}

					// RIGHT VALUE
					if ( SQStr::StrCmpi(watdata.wat_right.GetText(), TEXT("last")) == 0 )
					{
						right_value = left_value;
						if ( this->LeftOpRight(left_value, watdata.wat_op.GetText(), right_value) )
						{	// match
							this->WatchWarn(left_value, watdata, right_value);
						}
					} else if ( SQStr::StrCmpi(watdata.wat_right.GetText(), TEXT("prev")) == 0 )
					{
						int		ti;

						for ( ti = attrData->count - 1; ti >= 0; ti-- )
						{
							sdb::DBSmtData *	smtptr;
							int					dsec;

							smtptr = attrData->smtlist[ti];
							dsec = sdb::DiffTimeStamp(tsnow, smtptr->smt_ts);
							if ( dsec < this->opt.prev_time_sec )
								break;	// end
							if ( this->GetWatValue(right_value, watdata.wat_type, *smtptr) > 0 )
							{
								if ( this->LeftOpRight(left_value, watdata.wat_op.GetText(), right_value) )
								{	// match
									this->WatchWarn(left_value, watdata, right_value);
									break;
								}
							}
						}
					} else if ( SQStr::StrCmpi(watdata.wat_right.GetText(), TEXT("off")) == 0 )
					{	// 
					//	_tprintf(TEXT("right value off skip\n"));
						continue;
					} else
					{
						right_value = _ttoi64(watdata.wat_right.GetText());
						if ( this->LeftOpRight(left_value, watdata.wat_op.GetText(), right_value) )
						{	// match
							this->WatchWarn(left_value, watdata, right_value);
						}
					}

				} // end of for (wat_no;;)

				if ( this->opt.checkmode == CHECK_LONG )
				{	// NORMAL VALUE TEC
					this->AnalysAttributeValue(attr_id, *attrData, tsnow);
				}
				if ( attr_failer )
				{	// Failer
					this->warn[sdb::WATCHWARN_FAILURE]++;
				}

			} // end of for( attr_id;;)
		}
		this->WarnCheck(watdev.devdata);
		retcode = 0;
	} catch(int ret) { retcode = ret; }
	sqllib->ClearResult(sqlres);
	if ( smtlist != NULL )
		delete [] smtlist;
	if ( sdbdev != NULL )
		delete sdbdev;

	return retcode;
}

int		SdbWatch::WatchWarn(sdb::Int64 & left_value, sdb::DBWatData & watdata, sdb::Int64 right_value)
{
	TCHAR	tmp[256], t1[256];

	switch ( watdata.wat_type )
	{
	case sdb::WATCHTYPE_OFF:
		return 0;

	case sdb::WATCHTYPE_ATTR_VALUE:
		sdb::Fmt(t1, TEXT("Attribute value %I64u %s %I64u"), left_value, watdata.wat_op.GetDisplayText(), right_value );
		break;

	case sdb::WATCHTYPE_RAW_VALUE:
		sdb::Fmt(t1, TEXT("RAW value %I64u %s %I64u"), left_value, watdata.wat_op.GetDisplayText(), right_value );
		break;

	case sdb::WATCHTYPE_RAW_ANALYS:
		return 0;
	}
	this->WatchMessage(sdb::Fmt(tmp, TEXT("        Warnning: %s\n"), t1));
	if ( watdata.wat_warn >= 0 && watdata.wat_warn < sdb::WATCHWARN_NUM )
		this->warn[watdata.wat_warn]++;
	return 0;
}

void	SdbWatch::WarnClear(void)
{
	int		i;

	for ( i = 0; i < sdb::WATCHWARN_NUM; i++ )
		this->warn[i] = 0;
	return;
}

void	SdbWatch::WarnCheck(sdb::DBDevData & devdata)
{
	int		i, count = 0;

	for ( i = 0; i < sdb::WATCHWARN_NUM; i++ )
	{
		if ( i != sdb::WATCHWARN_OFF )
			count += this->warn[i];
	}
	if ( count > 0 )
	{
		TCHAR	tmp[256];

		sdb::Fmt(tmp, TEXT("Warnning: Failer %u, High %u, Mid %u, Low %u\n")
		, this->warn[sdb::WATCHWARN_FAILURE]
		, this->warn[sdb::WATCHWARN_HIGH   ]
		, this->warn[sdb::WATCHWARN_MID    ]
		, this->warn[sdb::WATCHWARN_LOW    ]
		);
		this->WatchMessage(tmp);

		if ( sdb::mail_warnlevel > 0 && this->opt.sendmail != FALSE )
		{
			BOOL	send_mail = FALSE;

			for ( i = 0; i < sdb::WATCHWARN_NUM; i++ )
			{
				if ( i != sdb::WATCHWARN_OFF && this->warn[i] > 0 )
				{
					if ( sdb::mail_warnlevel >= i )
					{
						send_mail = TRUE;
						break;
					}
				}
			}
			if ( send_mail )
			{
				this->watmsg.AddChar(0);
				this->SendMail(devdata, this->watmsg.GetText());
			}
		}
	}
	return;
}

int		SdbWatch::GetWatValue(sdb::Int64 & value, sdb::WatchType wat_type, sdb::DBSmtData & smtdata)
{
	switch ( wat_type )
	{
	case sdb::WATCHTYPE_OFF:
		return 0;

	case sdb::WATCHTYPE_ATTR_VALUE:
		value = smtdata.smt_value;
		break;

	case sdb::WATCHTYPE_RAW_VALUE:
		value = smtdata.smt_raw_value;
		break;

	case sdb::WATCHTYPE_RAW_ANALYS:
		return 0;
	}
	return 1;
}

BOOL	SdbWatch::LeftOpRight(sdb::Int64 left, const TCHAR * op, sdb::Int64 right)
{
	BOOL	f = FALSE;

	if ( SQStr::StrCmp(op, TEXT("==")) == 0 )
		f = left == right;
	else if ( SQStr::StrCmp(op, TEXT("!=")) == 0 )
		f = left != right;
	else if ( SQStr::StrCmp(op, TEXT(">=")) == 0 )
		f = left >= right;
	else if ( SQStr::StrCmp(op, TEXT("<=")) == 0 )
		f = left <= right;
	else if ( SQStr::StrCmp(op, TEXT(">")) == 0 )
		f = left > right;
	else if ( SQStr::StrCmp(op, TEXT("<")) == 0 )
		f = left < right;

	return f;
}

int		SdbWatch::AnalysAttributeValue(unsigned attr_id, SdbWatch::AttrData & attrData, SqlLib::TimeStamp & ts_base)
{
	RegAnalys			regAna1, regAna2;
	RegAnalys::XValue	x;
	RegAnalys::YValue	y, y1min, y1max, y2min, y2max;
	unsigned			i;

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"SdbWatch::AnalysAttributeValue(attr_id=%d)\n", attr_id));
	if ( regAna1.SetDataNum(attrData.count) < 0 )
		return -1;
	if ( regAna2.SetDataNum(attrData.count) < 0 )
		return -1;
	y1min = y1max = (RegAnalys::YValue) attrData.smtlist[0]->smt_value;
	y2min = y2max = (RegAnalys::YValue) attrData.smtlist[0]->smt_threshold;
	for ( i = 0; i < attrData.count; i++ )
	{
		x = (RegAnalys::XValue) sdb::DiffTimeStamp(attrData.smtlist[i]->smt_ts, ts_base);
		regAna1.SetXValue(i, x);
		regAna2.SetXValue(i, x);

		y = (RegAnalys::YValue) attrData.smtlist[i]->smt_value;
		if ( y1min > y )
			y1min = y;
		if ( y1max < y )
			y1max = y;
		regAna1.SetYValue(i, y);

		y = attrData.smtlist[i]->smt_threshold;
		if ( y2min > y )
			y2min = y;
		if ( y2max < y )
			y2max = y;
		regAna2.SetYValue(i, y);
	}
	if ( y1min == y1max && y2min == y2max )
		return 0;
	if ( regAna1.Analys() < 0 || regAna2.Analys() < 0 )
	{
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"  Analys error!\n"));
		return -1;
	}

	{
		RegAnalys::Value	a1, b1, a2, b2;
		RegAnalys::Value	sec, day;
		TCHAR				stt_msg[256], t1[256];
		SqlLib::TimeStamp	ts;

		a1 = regAna1.a;
		b1 = regAna1.b;
		a2 = regAna2.a;
		b2 = regAna2.b;
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL," a1=%f, b1=%f, a2=%f, b2=%f\n", (double)a1, (double)b1, (double)a2, (double)b2));
		if ( a1 != a2 )
		{
			sec =  (b2 - b1) / (a1 - a2);
			if ( sec > 0.0 )
			{
				day = sec / (RegAnalys::Value)(24 * 60 * 60);
				DBGREPORT((_CRT_WARN,NULL,NULL,NULL," day %d\n", (int)day));
				if ( day > 365.25 * 10 )	// 10 year over
				{
					sdb::Fmt(stt_msg, TEXT("over 10 years (%6.1f years)"), (double)(day/365.25));
				} else
				{
					sdb::OffsetTimeStamp(ts, ts_base, (int) sec);
					sdb::GetLocalTime(ts, ts);
					if ( day < 7 )
					{
						sdb::Fmt(stt_msg, TEXT("%04d-%02d-%02d %02d:%02d (%6.1f hours)")
						  , (int)ts.year, (int)ts.month, (int)ts.day, (int)ts.hour, (int)ts.minute
						  , (double) sec / 3600.0 );
					} else
					{
						sdb::Fmt(stt_msg, TEXT("%04d-%02d-%02d (%6.1f days)")
						  , (int)ts.year, (int)ts.month, (int)ts.day
						  , (double)day );
					}
					if ( day <= (RegAnalys::Value)sdb::warn_high_estimate )
						this->warn[sdb::WATCHWARN_HIGH]++;
					else if ( day <= (RegAnalys::Value)sdb::warn_mid_estimate )
						this->warn[sdb::WATCHWARN_MID]++;
					else if ( day <= (RegAnalys::Value)sdb::warn_low_estimate )
						this->warn[sdb::WATCHWARN_LOW]++;
				}
		//		this->WatchMessage(sdb::Fmt(t1, TEXT("        Threshold estimate %s\n"), stt_msg));
				PutPos(8, TEXT("Threshold"),
				      32, sdb::Fmt(t1, TEXT("estimate %s"    ), stt_msg  ),
				       0);
			}
		}
	}

	return 0;
}

int		SdbWatch::AnalysRawValue(unsigned attr_id, sdb::Int64 raw_value, SdbWatch::AttrData & attrData, SqlLib::TimeStamp & ts_base)
{
	RegAnalys			regAna;
	RegAnalys::XValue	x;
	RegAnalys::YValue	y, ymin, ymax;
	unsigned			i;

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"SdbWatch::AnalysRawValue(attr_id=%d, raw %I64d)\n", attr_id, raw_value));
	if ( regAna.SetDataNum(attrData.count) < 0 )
	{
		return -1;
	}
	ymin = ymax = (RegAnalys::YValue) attrData.smtlist[0]->smt_raw_value;
	for ( i = 0; i < attrData.count; i++ )
	{
		x = (RegAnalys::XValue) sdb::DiffTimeStamp(attrData.smtlist[i]->smt_ts, ts_base);
		regAna.SetXValue(i, x);
		y = (RegAnalys::YValue) attrData.smtlist[i]->smt_raw_value;
		if ( ymin > y )
			ymin = y;
		if ( ymax < y )
			ymax = y;
		regAna.SetYValue(i, y);
	}
	if ( ymin == ymax )
		return 0;
	if ( regAna.Analys() < 0 )
	{
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"  Analys error!\n"));
		return -1;
	}

	{
		RegAnalys::Value	a, b;
		RegAnalys::Value	sec, day;
		TCHAR				stt_msg[256], t1[256], t2[256];
		SqlLib::TimeStamp	ts;

		a = regAna.a;
		b = regAna.b;
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL," a=%f, b=%f\n", (double)a, (double)b));
		if ( a != 0.0 )
		{
			sec = ((RegAnalys::Value)raw_value - b) / a;
			if ( sec > 0.0 )
			{
				day = sec / (RegAnalys::Value)(24 * 60 * 60);
				DBGREPORT((_CRT_WARN,NULL,NULL,NULL," day %d\n", (int)day));
				if ( day > 365.25 * 10 )	// 10 year over
				{
					sdb::Fmt(stt_msg, TEXT("over 10 years (%6.1f years)"), (double)(day/365.25));
				} else
				{
					sdb::OffsetTimeStamp(ts, ts_base, (int) sec);
					sdb::GetLocalTime(ts, ts);
					if ( day < 7 )
					{
						sdb::Fmt(stt_msg, TEXT("%04d-%02d-%02d %02d:%02d (%6.1f hours)")
						  , (int)ts.year, (int)ts.month, (int)ts.day, (int)ts.hour, (int)ts.minute
						  , (double) sec / 3600.0 );
					} else
					{
						sdb::Fmt(stt_msg, TEXT("%04d-%02d-%02d (%6.1f days)")
						  , (int)ts.year, (int)ts.month, (int)ts.day
						  , (double)day );
					}
					if ( day <= (RegAnalys::Value)sdb::warn_high_estimate )
						this->warn[sdb::WATCHWARN_HIGH]++;
					else if ( day <= (RegAnalys::Value)sdb::warn_mid_estimate )
						this->warn[sdb::WATCHWARN_MID]++;
					else if ( day <= (RegAnalys::Value)sdb::warn_low_estimate )
						this->warn[sdb::WATCHWARN_LOW]++;
				}
			//	this->WatchMessage(sdb::Fmt(t1, TEXT("        RAW Value %I64d estimate %s\n"), raw_value, stt_msg));
				PutPos(8, sdb::Fmt(t1, TEXT("RAW Value %I64u"), raw_value),
				      32, sdb::Fmt(t2, TEXT("estimate %s"    ), stt_msg  ),
				       0);
			}
		}
	}

	return 0;
}

void	SdbWatch::PrintSmartData(sdb::DBDevData & devdata, sdb::DBSmtData & smtdata)
{
	TCHAR				att_name[80], att_msg[80], raw_msg[80], flg_msg[80], stt_msg[80];
	sdb::SmartVendor	vendor = sdb::SMART_VENDOR_GENERAL;
	TCHAR				t1[80];

	_tcsncpy(att_name, sdb::GetSmartAttributeIdName(vendor, smtdata.smt_attr_id), sizeof(att_name)/sizeof(*att_name) - 1);

	{
		unsigned	stt;

		stt = smtdata.smt_flags;
		sdb::Fmt(flg_msg, TEXT("%c%c%c%c%c%c")
		  , stt & 0x0020 ? TEXT('1') : TEXT('-')
		  , stt & 0x0010 ? TEXT('1') : TEXT('-')
		  , stt & 0x0008 ? TEXT('1') : TEXT('-')
		  , stt & 0x0004 ? TEXT('1') : TEXT('-')
		  , stt & 0x0002 ? TEXT('O') : TEXT('-')
		  , stt & 0x0001 ? TEXT('P') : TEXT('-')
		);
	}

	sdb::Fmt(att_msg, TEXT("%3u %3u %3u %s")
	           , smtdata.smt_value
	           , smtdata.smt_worst
	           , smtdata.smt_threshold
	           , flg_msg
	);

//	sdbdev->smart.attrList.GetSmartRawValue(*cur, sizeof(raw_msg)/sizeof(*raw_msg), raw_msg);
	sdb::Fmt(raw_msg, TEXT("%I64d"), (sdb::Int64)smtdata.smt_raw_value);

	if ( smtdata.smt_value > smtdata.smt_threshold )
		_tcscpy(stt_msg, TEXT("OK"));
	else
		_tcscpy(stt_msg, TEXT("FAILED"));

	PutPos(4     , sdb::Fmt(t1, TEXT("%3u %s"), smtdata.smt_attr_id, att_name),
	       4 + 4+sdb::SMART_ATTRIBUTE_MAX_LENGHTH   , att_msg,
	       COLUMN_MAX - 1 - 6 - 1 - _tcslen(raw_msg), raw_msg,
	       COLUMN_MAX - 1         - _tcslen(stt_msg), stt_msg,
	       0);
	return;
}

int		SdbWatch::WatchMessage(const TCHAR * text)
{
	int			ch;

	if ( text == NULL )
		return -1;
	_fputts(text, stdout);
	while ( (ch = *text++) != TEXT('\0') )
	{
		switch ( ch )
		{
		case TEXT('\r'):
			break;
		case TEXT('\n'):
			this->watmsg.AddText(TEXT("\r\n"));
			break;
		default:
			this->watmsg.AddChar(ch);
			break;
		}
	}
	return 0;
}

void	SdbWatch::PutPos(int pos, ... )
{
	va_list	arg;
	TCHAR	*text, tmp[256];;
	int		x = 0, len;

	va_start(arg, pos);
	while ( pos > 0 )
	{
		text = va_arg(arg, char*);
		len = (int)_tcslen(text);
		if ( pos > x )
		{
			this->WatchMessage(sdb::Fmt(tmp, TEXT("%*s"), pos - x, TEXT("")));
			x = pos;
		}
		this->WatchMessage(sdb::Fmt(tmp, TEXT("%s"), text));
		x += len;
		pos  = va_arg(arg, int);
	}
	if ( pos == 0 )
		this->WatchMessage(TEXT("\n"));
	va_end(arg);
}

int		SdbWatch::GetAttributeNum(AttrMap & attrmap, unsigned num, sdb::DBSmtData * smtlist)
{
	unsigned	i;
	unsigned	attr_id;
	unsigned	pos[SDB_SMART_ATTR_ID_MAX];

	attrmap.attr_num = 0;
	for ( attr_id = 0; attr_id < SDB_SMART_ATTR_ID_MAX; attr_id++ )
	{
		attrmap.id[attr_id].count = 0;
		pos[attr_id] = 0;
	}
	for ( i = 0; i < num; i++ )
	{
		attr_id = smtlist[i].smt_attr_id;
		if ( attr_id > 0 && attr_id < SDB_SMART_ATTR_ID_MAX )
			attrmap.id[attr_id].count++;
	}
	for ( attr_id = 0; attr_id < SDB_SMART_ATTR_ID_MAX; attr_id++ )
	{
		if ( attrmap.id[attr_id].count > 0 )
		{
			attrmap.attr_num++;
			if ( attrmap.id[attr_id].CreateList() < 0 )
				return -1;
		}
	}
	for ( i = 0; i < num; i++ )
	{
		unsigned	attr_id;

		attr_id = smtlist[i].smt_attr_id;
		if ( attr_id > 0 && attr_id < SDB_SMART_ATTR_ID_MAX )
		{
			_ASSERT(pos[attr_id] < attrmap.id[attr_id].count );
			attrmap.id[attr_id].smtlist[pos[attr_id]++] = &smtlist[i];
		}
	}
	for ( attr_id = 0; attr_id < SDB_SMART_ATTR_ID_MAX; attr_id++ )
	{
		_ASSERT(attrmap.id[attr_id].count == pos[attr_id] );
	}

	return (int)attrmap.attr_num;
}

//////////////////////////////////////////////////////////////////////////
//	SendMail

int		SdbWatch::SendMail(sdb::DBDevData & devdata, const TCHAR * body_text)
{
	SmtpLib		smtplib;
	int			retcode = 0;
	const TCHAR	*from_adr;
	const TCHAR	*to_adr;
	const TCHAR	*smtp_server;
	char		subject[1024];

	from_adr    = sdb::mail_from;
	to_adr      = sdb::mail_to;
	smtp_server = sdb::mail_smtp_server;
	if ( from_adr == NULL || to_adr == NULL || smtp_server == NULL )
		return -1;
	if ( strlen(from_adr) == 0 || strlen(to_adr) == 0 || strlen(smtp_server) == 0 )
		return -1;

	// Subject
	{
		if ( devdata.dev_hostname.GetLength() > 0 )
		{
			sprintf(subject, "SDBWATCH %s %s", devdata.dev_model.GetDisplayText(), devdata.dev_serial.GetDisplayText());
		} else
		{
			sprintf(subject, "SDBWATCH %s: %s %s", devdata.dev_hostname.GetDisplayText(), devdata.dev_model.GetDisplayText(), devdata.dev_serial.GetDisplayText());
		}
	}

	try
	{
		int				ret;
		VarBuffer		hbuf, bbuf;

		if ( smtplib.smtp_addr.SetAddr(smtp_server) < 0 )
			throw -1;
		smtplib.smtp_addr.SetPort(SmtpLib::PORT_SMTP);

		if ( smtplib.Create() < 0 )
			throw -1;

		// wb_쐬
		{
			char				tmp[4096];
			VarBuffer			vbuf;

			// From
			vbuf.Clear();
#if 1
			sprintf(tmp, "From: %s\r\n", from_adr);
#else
			SmtpLib::EncodeHeaderText(this->acc->smtpDisplayName.GetText(), vbuf);
			if ( (vbuf.GetSize() + this->acc->smtpEmail.GetLength()) * sizeof(char)  < sizeof(tmp) )
				sprintf(tmp, "From: %s <%s>\r\n", vbuf.GetText(), this->acc->smtpEmail.GetText() );
			else
				sprintf(tmp, "From: %s\r\n", this->acc->smtpEmail.GetText() );
#endif
			hbuf.AddText(tmp);

			// To
			vbuf.Clear();
			sprintf(tmp, "To: %s\r\n", to_adr);
			hbuf.AddText(tmp);

			// Subject
			vbuf.Clear();
			SmtpLib::MimeEncode(subject, vbuf);
			vbuf.AddChar(0);
			sprintf(tmp, "Subject: %s\r\n", vbuf.GetText());
			hbuf.AddText(tmp);

			// date
			{
				SqlLib::TimeStamp	st;
				SQStr				str;

				sdb::GetSystemTime(st);
				sprintf(tmp, "Date: %s\r\n", SdbWatch::GetDayTime(st, str));
				hbuf.AddText(tmp);
			}

		//	// X-Mailer
		//	sprintf(tmp, "X-Mailer: %s\r\n", YbbSendMail::YbbSendMail_XMailer);
		//	hbuf.AddText(tmp);
		}

		{
			const char	*header;

			hbuf.AddChar(0);
			header = hbuf.GetText();
			if ( (ret = smtplib.SendMail(from_adr, to_adr, body_text, header, TRUE)) < 0 )
				throw -1;
		}

		retcode = 0;
	} catch(int ret) { retcode = ret; }
	if ( smtplib.IsCreate() )
		smtplib.Close();

	return retcode;
}

const TCHAR *	SdbWatch::GetDayTime(const SqlLib::TimeStamp & st, SQStr & str)
{
	static const TCHAR	*montbl[12] =
	{
		TEXT("Jan"), TEXT("Feb"), TEXT("Mar"), TEXT("Apr"), TEXT("May"), TEXT("Jun"),
		TEXT("Jul"), TEXT("Aug"), TEXT("Sep"), TEXT("Oct"), TEXT("Nov"), TEXT("Dec"),
	};
	SqlLib::TimeStamp	lt;
	TCHAR				tmp[256], zone[256];

	sdb::GetLocalTime(lt, st);
#if	defined(WIN32)
	{
		DWORD					dw;
		TIME_ZONE_INFORMATION	tzinf;

		dw = ::GetTimeZoneInformation(&tzinf);
		switch ( dw )
		{
		case TIME_ZONE_ID_UNKNOWN:
		case TIME_ZONE_ID_STANDARD:
		case TIME_ZONE_ID_DAYLIGHT:
			_stprintf(zone, TEXT("%+02d%02u")
			  , (int)(-tzinf.Bias) / 60
			  , abs((-tzinf.Bias) % 60)
			);
			break;
		case TIME_ZONE_ID_INVALID:
		default:
			_stprintf(zone, TEXT("UT"));
			break;
		}
	}
#else
	{
		int		tz_min;

		tz_min = timezone / 60;
		_stprintf(zone, TEXT("%+02d%02u")
		  , (int)(tz_min) / 60
		  , abs((tz_min) % 60)
		);
	}
#endif
	_stprintf(tmp, TEXT("%u %s %u %02u:%02u:%02u %s")
	  , (unsigned)lt.day
	  , montbl[(lt.month - 1) % 12]
	  , (unsigned)lt.year
	  , (unsigned)lt.hour
	  , (unsigned)lt.minute
	  , (unsigned)lt.second
	  , zone
	);
	str.SetText(tmp);
	return str.GetText();
}

//////////////////////////////////////////////////////////////////////////
//	WatchDev
//////////////////////////////////////////////////////////////////////////

SdbWatch::WatchDev::WatchDev()
{
}

SdbWatch::WatchDev::~WatchDev()
{
}

//////////////////////////////////////////////////////////////////////////
//	AttrData
//////////////////////////////////////////////////////////////////////////

SdbWatch::AttrData::AttrData()
{
	unsigned	i;

	this->count = 0;
	this->smtlist = NULL;
	for ( i = 0; i < sizeof(this->ww)/sizeof(*this->ww); i++ )
	{
		this->ww[i].count     = 0;
		this->ww[i].watdata.wat_attr_id = 0;
	}
	return;
}

SdbWatch::AttrData::~AttrData()
{
	if ( this->smtlist != NULL )
		delete [] this->smtlist;
	return;
}

int		SdbWatch::AttrData::CreateList(void)
{
	if ( this->count == 0 )
		return -1;
	if ( this->smtlist != NULL )
		delete [] this->smtlist;
	if ( (this->smtlist = new sdb::DBSmtData * [this->count]) == NULL )
		return -1;
	return 0;
}


